
import threading

from .localrunner import WorkflowRunner
from .workflow_widget import WorkflowWidget
from .monitor_widget import MonitorWidget, update_monitor


def workflow_widget(**kwargs):
    """
    Creates and displays a widget for workflow definitions. Passes any given
    keyword arguments to the WorkflowWidget constructor.

    :return: (function call to 'WorkflowWidget.display_widget)
    """

    widget = WorkflowWidget(**kwargs)

    return widget.display_widget()


def monitor_widget(**kwargs):
    """
    Creates and displays a widget for monitoring Vgrid job queues. Passes
    any given keyword arguments to the MonitorWidget constructor.

    :return: (function call to 'MonitorWidget.display_widget)
    """

    widget = MonitorWidget(**kwargs)

    monitor_thread = threading.Thread(
        target=update_monitor,
        args=(widget,),
        daemon=True
    )

    monitor_thread.start()

    return widget.display_widget()


def start_local_workflow(
        vgrid_name, patterns, recipes, workers, warning=True,
        print_logging=True, start_workers=True, daemon=False, settle_time=3):
    if warning:
        print("This function is intended only as an illustration of MEOW "
              "functionality, and should therefore be used with caution. It "
              "may behave unexpectedly, is not thread safe, has no protection "
              "for data being overridden, and very little error handling. For "
              "proper workflow functionality please use the MiG integration")

    runner = WorkflowRunner(
        vgrid_name,
        workers,
        patterns=patterns,
        recipes=recipes,
        start_workers=start_workers,
        print_logging=print_logging,
        daemon=daemon,
        settle_time=settle_time
    )

    return runner



from .version import __name__, __version__, __fullname__


def info():
    """
    Debug function to check that mig_meow has been imported correctly.
    Prints message about the current build.
    
    :return: (str) debug message.  
    """

    msg = 'ver: %s\n' \
          '%s has been imported correctly. \n' \
          '%s is a paradigm used for defining event based ' \
          'workflows. It is designed primarily to work with IDMC, a MiG ' \
          'implementation available at the University of Copenhagen. ' \
          % (__version__, __name__, __fullname__)
    print(msg)




# This code is heavily based off the 'mig/server/grid_events.py' file
# contained in the MiG source code at: https://sourceforge.net/projects/migrid/

import copy
import glob
import os
import time
import shutil
import re
import fnmatch
import threading
import multiprocessing

from .constants import PATTERNS, RECIPES, NAME, SOURCE, CHAR_LOWERCASE, \
    CHAR_UPPERCASE, CHAR_NUMERIC, RECIPE, KEYWORD_DIR, KEYWORD_EXTENSION, \
    KEYWORD_FILENAME, KEYWORD_JOB, KEYWORD_PATH, KEYWORD_PREFIX, \
    KEYWORD_REL_DIR, KEYWORD_REL_PATH, KEYWORD_VGRID, VGRID
from .logging import create_localrunner_logfile, write_to_log
from .fileio import write_dir_pattern, write_dir_recipe, make_dir, \
    read_dir_recipe, read_dir_pattern, write_notebook, write_yaml, read_yaml, \
    delete_dir_pattern, delete_dir_recipe
from .meow import get_parameter_sweep_values, is_valid_pattern_object, Pattern
from .validation import valid_dir_path, check_input, is_valid_recipe_dict

from datetime import datetime
from random import SystemRandom
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler, FileCreatedEvent, \
    FileModifiedEvent, FileDeletedEvent, DirCreatedEvent, DirModifiedEvent, \
    DirDeletedEvent

rule_hits = {}
_hits_lock = threading.Lock()

(_rate_limit_field, _settle_time_field) = ('rate_limit', 'settle_time')
_default_period = 'm'
_default_time = '0'
_unit_periods = {
    's': 1,
    'm': 60,
    'h': 60 * 60,
    'd': 24 * 60 * 60,
    'w': 7 * 24 * 60 * 60,
}
DEFAULT_SETTLE_TIME = 3
SETTLE_TIME = 'settle_time'

_trigger_event = '_trigger_event'

RUNNER_DATA = '.workflow_runner_data'
OUTPUT_DATA = 'job_output'

JOB_DIR = '.workflow_processing'

LOGGER = 'logger'
RULES = 'rules'
RULE_ID = 'id'
RULE_PATH = 'path'
RULE_PATTERN = 'pattern'
RULE_RECIPE = 'recipe'
ADMIN = 'admin'
MONITOR = 'monitor'
WORKERS = 'workers'
JOBS = 'jobs'
QUEUE = 'queue'
DATA_DIR = 'data_dir'
JOBS_DIR = 'jobs_dir'
OUTPUT_DIR = 'output_dir'
RETRO_ACTIVE = 'retro'
PRINT = 'print'

QUEUED = 'queued'
RUNNING = 'running'
FAILED = 'failed'
DONE = 'done'

JOB_ID = 'id'
JOB_PATTERN = 'pattern'
JOB_RECIPE = 'recipe'
JOB_RULE = 'rule'
JOB_PATH = 'path'
JOB_STATUS = 'status'
JOB_CREATE_TIME = 'create'
JOB_START_TIME = 'start'
JOB_END_TIME = 'end'
JOB_ERROR = 'error'

META_FILE = 'job.yml'
BASE_FILE = 'base.ipynb'
PARAMS_FILE = 'params.yml'
JOB_FILE = 'job.ipynb'
RESULT_FILE = 'result.ipynb'

recent_jobs = {}
_recent_jobs_lock = threading.Lock()
_job_lock = threading.Lock()
_queue_lock = threading.Lock()
_worker_lock = threading.Lock()
_rules_lock = threading.Lock()


def get_runner_patterns(runner_data):
    return os.path.join(runner_data, PATTERNS)


def get_runner_recipes(runner_data):
    return os.path.join(runner_data, RECIPES)


def generate_id(length=16):
    """
    Generates a random id by using randomly generated alphanumeric strings.
    Uniqueness is not guaranteed, but is a reasonable assumption.

    :param length: (int) [optional] The length of the id to be generated.
    Default is 16

    :return: (str) A random collection of alphanumeric characters.
    """
    charset = CHAR_UPPERCASE + CHAR_LOWERCASE + CHAR_NUMERIC
    return ''.join(SystemRandom().choice(charset) for _ in range(length))


def make_fake_event(path, state, is_directory=False):
    """Create a fake state change event for path. Looks up path to see if the
    change is a directory or file.
    """

    file_map = {'modified': FileModifiedEvent,
                'created': FileCreatedEvent,
                'deleted': FileDeletedEvent}
    dir_map = {'modified': DirModifiedEvent,
               'created': DirCreatedEvent, 'deleted': DirDeletedEvent}
    if is_directory or os.path.isdir(path):
        fake = dir_map[state](path)
    else:
        fake = file_map[state](path)

    # mark it a trigger event
    setattr(fake, _trigger_event, True)
    return fake


def is_fake_event(event):
    """Check if event came from our trigger-X rules rather than a real file
    system change.
    """

    return getattr(event, _trigger_event, False)


def get_job_dir(state, job_id):
    """
    Gets the path of a jobs directory.

    :param state: (dict) The shared state.

    :param job_id: (str) The job id.

    :return: (str) The path to the jobs directory.
    """
    return os.path.join(state[JOBS_DIR], job_id)


def replace_keywords(old_dict, state, job_id, src_path):
    """
    Replaces MiG trigger keywords with with actual values.

    :param old_dict: (dict) A values dict potentially containing MiG keywords.

    :param state: (dict) The shared runner state

    :param job_id: (str) The appropriate job ID, corresponding to old_dict.

    :param src_path: (str) The triggering path for the event generating the
    job_id job

    :return: (dict) A dict corresponding to old_dict, with the MiG keywords
    replaced with appropriate values.
    """
    new_dict = {}

    filename = os.path.basename(src_path)
    dirname = os.path.dirname(src_path)
    relpath = os.path.relpath(src_path, state[VGRID])
    reldirname = os.path.dirname(relpath)
    (prefix, extension) = os.path.splitext(filename)

    for var, val in old_dict.items():
        if isinstance(val, str):
            val = val.replace(KEYWORD_PATH, src_path)
            val = val.replace(KEYWORD_REL_PATH, relpath)
            val = val.replace(KEYWORD_DIR, dirname)
            val = val.replace(KEYWORD_REL_DIR, reldirname)
            val = val.replace(KEYWORD_FILENAME, filename)
            val = val.replace(KEYWORD_PREFIX, prefix)
            val = val.replace(KEYWORD_VGRID, state[VGRID])
            val = val.replace(KEYWORD_EXTENSION, extension)
            val = val.replace(KEYWORD_JOB, job_id)

            new_dict[var] = val
        else:
            new_dict[var] = val

    return new_dict


def schedule_job(runner_state, rule, src_path, recipe_code, yaml_dict):
    """
    Schedules a new job in the workflow runner. This creates the appropriate
    job files in a shared directory, adds the job to the queue, and add it to
    the list of all jobs.

    :param runner_state: (dict) The shared runner state.

    :param rule: (dict) The rule causing this job to be scheduled.

    :param src_path: (str) The path which generated the triggering event.

    :param recipe_code: (dict) The recipe code to be run through.

    :param yaml_dict: (dict) Any variables to be applied.

    :return: No return.
    """
    job_dict = {
        JOB_ID: generate_id(),
        JOB_PATTERN: rule[RULE_PATTERN],
        JOB_RECIPE: rule[RULE_RECIPE],
        JOB_RULE: rule[RULE_ID],
        JOB_PATH: src_path,
        JOB_STATUS: QUEUED,
        JOB_CREATE_TIME: datetime.now()
    }

    yaml_dict = replace_keywords(
        yaml_dict,
        runner_state,
        job_dict[JOB_ID],
        src_path
    )

    print('edited yaml dict to: %s' % yaml_dict)

    job_dir = get_job_dir(runner_state, job_dict[JOB_ID])
    make_dir(job_dir)

    meta_file = os.path.join(job_dir, META_FILE)
    write_yaml(job_dict, meta_file)

    base_file = os.path.join(job_dir, BASE_FILE)
    write_notebook(recipe_code, base_file)

    yaml_file = os.path.join(job_dir, PARAMS_FILE)
    write_yaml(yaml_dict, yaml_file)

    _job_lock.acquire()
    _queue_lock.acquire()
    try:
        runner_state[JOBS].append(job_dict[JOB_ID])
        runner_state[QUEUE].append(job_dict[JOB_ID])
    except Exception as ex:
        _job_lock.release()
        _queue_lock.release()
        raise Exception(ex)
    _job_lock.release()
    _queue_lock.release()


def runner_log(runner_state, func_name, log_msg):
    # queue_count = len(runner_state[QUEUE])
    # jobs_count = len(runner_state[JOBS])

    # status_msg = f"({queue_count}, {jobs_count}) {log_msg}"
    status_msg = func_name + ': ' +log_msg

    write_to_log(
        runner_state[LOGGER],
        func_name,
        status_msg
    )

    if runner_state[PRINT]:
        print(status_msg)
        # print('\r' + status_msg, end='')


class WorkflowRunner:
    def __init__(self, path, workers, patterns=None, recipes=None,
                 meow_data=RUNNER_DATA, job_data=JOB_DIR,
                 output_data=OUTPUT_DATA, daemon=False, reuse_vgrid=True,
                 start_workers=False, retro_active_jobs=True,
                 print_logging=True, file_logging=False,
                 settle_time=DEFAULT_SETTLE_TIME):

        valid_dir_path(path, 'path')
        check_input(workers, int, 'workers', or_none=True)
        check_input(patterns, dict, PATTERNS, or_none=True)
        check_input(recipes, dict, RECIPES, or_none=True)
        valid_dir_path(meow_data, 'meow_data')
        valid_dir_path(job_data, 'job_data')
        valid_dir_path(output_data, 'output_data')
        check_input(daemon, bool, 'daemon')
        check_input(reuse_vgrid, bool, 'reuse_vgrid')
        check_input(start_workers, bool, 'start_workers')
        check_input(retro_active_jobs, bool, 'retro_active_jobs')
        check_input(print_logging, bool, 'print_logging')
        check_input(file_logging, bool, 'file_logging')
        check_input(settle_time, int, 'settle_time')

        self.__runner_state = {
            PATTERNS: {},
            RECIPES: {},
            RULES: [],
            JOBS: [],
            QUEUE: [],
            LOGGER: None,
            ADMIN: None,
            MONITOR: None,
            WORKERS: [],
            DATA_DIR: None,
            JOBS_DIR: None,
            OUTPUT_DIR: None,
            VGRID: None,
            RETRO_ACTIVE: retro_active_jobs,
            PRINT: print_logging,
            SETTLE_TIME: settle_time
        }

        runner_log_file = create_localrunner_logfile(debug_mode=file_logging)
        self.__runner_state[LOGGER] = runner_log_file
        runner_log(
            self.__runner_state,
            'run_local_workflow',
            'created new log at %s' % self.__runner_state[LOGGER]
        )

        make_dir(path, can_exist=reuse_vgrid)
        make_dir(job_data)
        if meow_data == RUNNER_DATA:
            make_dir(meow_data, ensure_clean=True)
        else:
            make_dir(meow_data)
        make_dir(output_data)
        make_dir(get_runner_patterns(meow_data), ensure_clean=True)
        make_dir(get_runner_recipes(meow_data), ensure_clean=True)

        self.__runner_state[VGRID] = path

        runner_log(
            self.__runner_state,
            'run_local_workflow',
            'Starting MEOW monitor'
        )

        self.__runner_state[DATA_DIR] = meow_data
        self.__runner_state[JOBS_DIR] = job_data
        self.__runner_state[OUTPUT_DIR] = output_data

        workflow_administrator = LocalWorkflowAdministrator(
            runner_state=self.__runner_state
        )
        administrator_observer = Observer()
        self.__runner_state[ADMIN] = administrator_observer
        administrator_observer.schedule(
            workflow_administrator,
            meow_data,
            recursive=True
        )
        administrator_observer.start()

        runner_log(
            self.__runner_state,
            'run_local_workflow',
            'Starting file monitor'
        )

        workflow_monitor = LocalWorkflowMonitor(
            runner_state=self.__runner_state
        )
        monitor_observer = Observer()
        self.__runner_state[MONITOR] = monitor_observer
        monitor_observer.schedule(
            workflow_monitor,
            path,
            recursive=True
        )
        monitor_observer.start()

        runner_log(
            self.__runner_state,
            'run_local_workflow',
            'Monitor setup complete'
        )

        if patterns:
            for name, pattern in patterns.items():
                write_dir_pattern(pattern, directory=meow_data)

        if recipes:
            for name, recipe in recipes.items():
                write_dir_recipe(recipe, directory=meow_data)

        runner_log(
            self.__runner_state,
            'run_local_workflow',
            'Initial Pattern and Recipe definitions complete'
        )

        _worker_lock.acquire()
        try:
            for id in range(0, workers):
                worker = JobProcessor(id, self.__runner_state)
                self.__runner_state[WORKERS].append(worker)
        except Exception as ex:
            _worker_lock.release()
            raise Exception(ex)
        _worker_lock.release()

        if start_workers:
            self.start_workers()

        if not daemon:
            try:
                while True:
                    time.sleep(1)
            except KeyboardInterrupt:
                self.stop_runner()

    def start_workers(self):
        _worker_lock.acquire()
        try:
            workers = self.__runner_state[WORKERS]

            for worker in workers:
                runner_log(
                    self.__runner_state,
                    'start_workers',
                    "Starting worker %s" % worker.worker_id
                )

                worker.start()
        except Exception as ex:
            _worker_lock.release()
            raise Exception(ex)
        _worker_lock.release()

    def check_running_status(self):
        if not self.__runner_state[ADMIN]:
            return (False, 'The Workflow Admin is not running. You should '
                           'start another workflow runner. ')
        if not self.__runner_state[MONITOR]:
            return (False, 'The Workflow Monitor is not running. You should '
                           'start another workflow runner. ')
        running, total = self.get_running_status()
        if running == total:
            return (True, 'All workers are running. ')
        else:
            return (False, '%d workers are not running. ' % (total - running))

    def get_running_status(self):
        _worker_lock.acquire()
        total = len(self.__runner_state[WORKERS])
        running = 0
        try:
            for worker in self.__runner_state[WORKERS]:
                if worker.isAlive():
                    running += 1
        except Exception as ex:
            _worker_lock.release()
            raise Exception(ex)
        _worker_lock.release()
        return (running, total)

    def stop_runner(self, clear_jobs=False):
        _worker_lock.acquire()
        try:
            monitor_observer = self.__runner_state[MONITOR]
            administrator_observer = self.__runner_state[ADMIN]
            workers = self.__runner_state[WORKERS]

            to_join = []

            if monitor_observer.is_alive():
                to_join.append(monitor_observer)
                monitor_observer.stop()

            if administrator_observer.is_alive():
                to_join.append(administrator_observer)
                administrator_observer.stop()
            for worker in workers:
                if worker.is_alive():
                    to_join.append(worker)
                    worker.stop()

            for thread in to_join:
                thread.join()

            self.__runner_state[MONITOR] = None
            self.__runner_state[ADMIN] = None
            self.__runner_state[WORKERS] = []
        except Exception as ex:
            _worker_lock.release()
            raise Exception(ex)
        _worker_lock.release()

        meow_data = self.__runner_state[DATA_DIR]
        if os.path.exists(meow_data) \
                and os.path.isdir(meow_data) \
                and meow_data == RUNNER_DATA:
            shutil.rmtree(meow_data)

        job_data = self.__runner_state[JOBS_DIR]
        if clear_jobs and os.path.exists(job_data):
            for job in self.__runner_state[JOBS]:
                job_dir = get_job_dir(self.__runner_state, job)
                if os.path.exists(job_dir):
                    shutil.rmtree(job_dir)
            if len(os.listdir(job_data)) == 0:
                os.rmdir(job_data)
        return True

    def get_all_jobs(self):
        job_queue = []

        _job_lock.acquire()
        try:
            job_queue = copy.deepcopy(self.__runner_state[JOBS])
        except:
            pass
        _job_lock.release()

        return job_queue

    def get_queued_jobs(self):
        job_queue = []

        _job_lock.acquire()
        try:
            job_queue = copy.deepcopy(self.__runner_state[QUEUE])
        except:
            pass
        _job_lock.release()

        return job_queue

    def get_all_input_paths(self):
        input_paths = []
        _rules_lock.acquire()
        try:
            for rule in self.__runner_state[RULES]:
                input_paths.append(rule[RULE_PATH])
            _rules_lock.release()
        except Exception as ex:
            _rules_lock.release()
            raise Exception(ex)
        return  input_paths

    def check_status(self):
        queued_jobs = len(self.get_queued_jobs())
        all_jobs = len(self.get_all_jobs())
        input_paths = self.get_all_input_paths()

        status = f"[{queued_jobs}-{all_jobs}] () {input_paths}"

        return status

    def add_pattern(self, pattern):
        status, msg = is_valid_pattern_object(pattern)
        if not status:
            runner_log(
                self.__runner_state,
                'add_pattern',
                'Could not add pattern. %s' % msg
            )
        else:
            write_dir_pattern(pattern, directory=self.__runner_state[DATA_DIR])

            runner_log(
                self.__runner_state,
                'add_pattern',
                'Added pattern: %s' % msg
            )

    def modify_pattern(self, pattern):
        status, msg = is_valid_pattern_object(pattern)
        if not status:
            runner_log(
                self.__runner_state,
                'modify_pattern',
                'Could not modify pattern. %s' % msg
            )
        else:
            write_dir_pattern(pattern, directory=self.__runner_state[DATA_DIR])

            runner_log(
                self.__runner_state,
                'modify_pattern',
                'Modified pattern: %s' % msg
            )

    def remove_pattern(self, pattern):
        if isinstance(pattern, str):
            name = pattern
        elif isinstance(pattern, Pattern):
            name = pattern.name
        else:
            runner_log(
                self.__runner_state,
                'remove_pattern',
                'Invalid pattern parameter. Must be either Pattern object, '
                'or a str name. '
            )
            return

        delete_dir_pattern(name, directory=self.__runner_state[DATA_DIR])
        runner_log(
            self.__runner_state,
            'remove_pattern',
            'Removed pattern: %s' % name
        )

    def add_recipe(self, recipe):
        status, msg = is_valid_recipe_dict(recipe)
        if not status:
            runner_log(
                self.__runner_state,
                'add_recipe',
                'Could not add recipe. %s' % msg
            )
        else:
            write_dir_recipe(recipe, directory=self.__runner_state[DATA_DIR])

            runner_log(
                self.__runner_state,
                'add_recipe',
                'Added user recipe: %s' % msg
            )

    def modify_recipe(self, recipe):
        status, msg = is_valid_recipe_dict(recipe)
        if not status:
            runner_log(
                self.__runner_state,
                'modify_recipe',
                'Could not modify recipe. %s' % msg
            )
        else:
            write_dir_recipe(recipe, directory=self.__runner_state[DATA_DIR])

            runner_log(
                self.__runner_state,
                'modify_recipe',
                'Modified recipe: %s' % msg
            )

    def remove_recipe(self, recipe):
        if isinstance(recipe, str):
            name = recipe
        elif isinstance(recipe, dict):
            name = recipe[NAME]
        else:
            runner_log(
                self.__runner_state,
                'remove_recipe',
                'Invalid recipe parameter. Must be either recipe dict, '
                'or a str name. '
            )
            return

        delete_dir_recipe(name, directory=self.__runner_state[DATA_DIR])
        runner_log(
            self.__runner_state,
            'remove_recipe',
            'Removed recipe: %s' % name
        )

    def check_recipes(self):
        recipes_dict = copy.deepcopy(self.__runner_state[RECIPES])
        return recipes_dict

    def check_patterns(self):
        patterns_dict = copy.deepcopy(self.__runner_state[PATTERNS])
        return patterns_dict

    def check_rules(self):
        rules_dict = copy.deepcopy(self.__runner_state[RULES])
        return rules_dict

    def check_jobs(self):
        jobs_dict = copy.deepcopy(self.__runner_state[JOBS])
        return jobs_dict

    def check_queue(self):
        jobs_dict = copy.deepcopy(self.__runner_state[QUEUE])
        return jobs_dict


class JobProcessor(threading.Thread):
    def __init__(self, worker_id, runner_state):
        threading.Thread.__init__(self)
        self.worker_id = worker_id
        self._stop_flag = threading.Event()
        self.runner_state = runner_state

    def run(self):
        try:
            while True:
                if self._stop_flag.isSet():
                    return
                _queue_lock.acquire()
                try:
                    queue = self.runner_state[JOBS]

                    # runner_log(
                    #     self.runner_state,
                    #     'worker %s' % self.worker_id,
                    #     "There are %d jobs in the queue" % len(queue)
                    # )

                    running_job = None
                    running_data = None
                    for job in queue:
                        job_dir = get_job_dir(self.runner_state, job)
                        meta_path = os.path.join(job_dir, META_FILE)

                        job_data = read_yaml(meta_path)

                        if job_data[JOB_STATUS] == QUEUED:
                            running_job = job
                            running_data = job_data
                            job_data[JOB_STATUS] = RUNNING
                            job_data[JOB_START_TIME] = datetime.now()

                            write_yaml(job_data, meta_path)

                            runner_log(
                                self.runner_state,
                                'worker %s' % self.worker_id,
                                "Found job %s" % job_data[JOB_ID]
                            )
                            break
                except Exception as ex:
                    _queue_lock.release()
                    raise Exception(ex)
                if running_job:
                    self.runner_state[JOBS].remove(running_job)
                    _queue_lock.release()
                else:
                    _queue_lock.release()
                    time.sleep(10 + (self.worker_id % 10))
                    continue

                job_dir = get_job_dir(self.runner_state, running_job)
                meta_path = os.path.join(job_dir, META_FILE)
                base_path = os.path.join(job_dir, BASE_FILE)
                param_path = os.path.join(job_dir, PARAMS_FILE)
                job_path = os.path.join(job_dir, JOB_FILE)
                result_path = os.path.join(job_dir, RESULT_FILE)

                error = False
                cmd = 'notebook_parameterizer ' \
                      + base_path + ' ' \
                      + param_path + ' ' \
                      + '-o ' + job_path
                try:
                    os.system(cmd)
                except Exception as ex:
                    error = ex

                if not os.path.exists(job_path) or error:
                    running_data[JOB_STATUS] = FAILED
                    running_data[JOB_END_TIME] = datetime.now()
                    msg = 'Job file %s was not created successfully' \
                          % running_job
                    if error:
                        msg += '. %s' % error
                    running_data[JOB_ERROR] = msg
                    write_yaml(running_data, meta_path)
                    runner_log(
                        self.runner_state,
                        'worker %s' % self.worker_id,
                        "Job worker encountered and error. %s" % msg
                    )
                    time.sleep(10 + (self.worker_id % 10))
                    continue

                cmd = 'papermill ' \
                      + job_path + ' ' \
                      + result_path
                try:
                    os.system(cmd)
                except Exception as ex:
                    error = ex

                if not os.path.exists(result_path) or error:
                    running_data[JOB_STATUS] = FAILED
                    running_data[JOB_END_TIME] = datetime.now()
                    msg = 'Result file %s was not created successfully'
                    if error:
                        msg += '. %s' % error
                    running_data[JOB_ERROR] = msg
                    write_yaml(running_data, meta_path)
                    runner_log(
                        self.runner_state,
                        'worker %s' % self.worker_id,
                        "Job worker encountered and error. %s" % msg
                    )
                    time.sleep(10 + (self.worker_id % 10))
                    continue

                running_data[JOB_STATUS] = DONE
                running_data[JOB_END_TIME] = datetime.now()
                write_yaml(running_data, meta_path)

                job_output_dir = \
                    os.path.join(self.runner_state[OUTPUT_DIR], running_job)

                shutil.copytree(job_dir, job_output_dir)

                time.sleep(10 + (self.worker_id % 10))
        except Exception as ex:
            runner_log(
                self.runner_state,
                'worker %s' % self.worker_id,
                "Worker has crashed. %s" % str(ex)
            )
            raise Exception(ex)

    def stop(self):
        self._stop_flag.set()


class LocalWorkflowAdministrator(PatternMatchingEventHandler):
    """
    Event handler to monitor pattern and recipe changes.
    """

    def __init__(
            self, runner_state=None, patterns=None, ignore_patterns=None,
            ignore_directories=False, case_sensitive=False):
        """Constructor"""

        PatternMatchingEventHandler.__init__(
            self,
            patterns,
            ignore_patterns,
            ignore_directories,
            case_sensitive
        )
        self.runner_state = runner_state

        # Check pre-existing Patterns and Recipes
        runner_patterns = get_runner_patterns(runner_state[DATA_DIR])
        for file_path in os.listdir(runner_patterns):
            try:
                pattern = read_dir_pattern(
                    file_path,
                    directory=self.runner_state[DATA_DIR]
                )
                self.add_pattern(pattern)
            except Exception as exc:
                runner_log(
                    self.runner_state,
                    'LocalWorkflowAdministrator._init',
                    str(exc)
                )

        runner_recipes = get_runner_recipes(runner_state[DATA_DIR])
        for file_path in os.listdir(runner_recipes):
            try:
                recipe = read_dir_recipe(
                    file_path,
                    directory=self.runner_state[DATA_DIR]
                )
                self.add_recipe(recipe)
            except Exception as exc:
                runner_log(
                    self.runner_state,
                   'LocalWorkflowAdministrator._init',
                    str(exc)
                )

    def update_rules(self, event):
        """Handle all rule updates"""

        if event.is_directory:
            return

        runner_log(
            self.runner_state,
            'update_rules',
            "Handling %s rule update at %s"
            % (event.event_type, event.src_path)
        )

        src_path = event.src_path

        try:
            valid_dir_path(src_path, 'src_path')
        except ValueError as ve:
            runner_log(
                self.runner_state,
                'update_rules',
                "Ignoring events at %s. %s" % (src_path, ve)
            )
            return

        event_type = event.event_type
        file_type = ''
        file_path = ''
        try:
            runner_patterns = get_runner_patterns(self.runner_state[DATA_DIR])
            runner_recipes = get_runner_recipes(self.runner_state[DATA_DIR])
            if runner_patterns in src_path:
                file_path = src_path[
                            src_path.find(runner_patterns)
                            + len(runner_patterns)+1:]
                file_type = PATTERNS
            elif runner_recipes in src_path:
                file_path = src_path[
                            src_path.find(runner_recipes)
                            + len(runner_recipes)+1:]
                file_type = RECIPES
        except Exception as exc:
            runner_log(
                self.runner_state,
                'update_rules-pattern',
                'Cannot process event at %s due to error: %s'
                % (src_path, exc)
            )
            return
        if os.path.sep in file_path:
            runner_log(
                self.runner_state,
                'update_rules-pattern',
                'Cannot process nested event at %s' % src_path
            )
            return

        if event_type in ['created', 'modified']:
            if file_type == PATTERNS:
                try:
                    pattern = read_dir_pattern(
                        file_path,
                        directory=self.runner_state[DATA_DIR]
                    )
                except Exception as exc:
                    runner_log(
                        self.runner_state,
                        'update_rules-pattern',
                        str(exc)
                    )
                    return
                self.add_pattern(pattern)
            elif file_type == RECIPES:
                try:
                    recipe = read_dir_recipe(
                        file_path,
                        directory=self.runner_state[DATA_DIR]
                    )
                except Exception as exc:
                    runner_log(
                        self.runner_state,
                        'update_rules-recipe',
                        str(exc)
                    )
                    return
                self.add_recipe(recipe)
        elif event_type == 'deleted':
            if file_type == PATTERNS:
                self.remove_pattern(file_path)
            elif file_type == RECIPES:
                self.remove_recipe(file_path)

    def on_modified(self, event):
        """Handle modified rule file"""

        self.update_rules(event)

    def on_created(self, event):
        """Handle new rule file"""

        self.update_rules(event)

    def on_deleted(self, event):
        """Handle deleted rule file"""

        self.update_rules(event)

    def add_pattern(self, pattern):
        op = 'Created new'
        if pattern.name in self.runner_state[PATTERNS]:
            if self.runner_state[PATTERNS][pattern.name] == pattern:
                return
            else:
                self.remove_pattern(pattern.name)
                op = 'Modified'
        self.runner_state[PATTERNS][pattern.name] = pattern
        self.identify_rules(new_pattern=pattern)
        runner_log(
            self.runner_state,
            'add_pattern',
            '%s pattern %s' % (op, pattern)
        )

    def add_recipe(self, recipe):
        op = 'Created new'
        if recipe[NAME] in self.runner_state[RECIPES]:
            if self.runner_state[RECIPES][recipe[NAME]] == recipe:
                return
            else:
                self.remove_recipe(recipe[NAME])
                op = 'Modified'
        self.runner_state[RECIPES][recipe[NAME]] = recipe
        self.identify_rules(new_recipe=recipe)
        runner_log(
            self.runner_state,
            'add_recipe',
            '%s recipe %s from source %s'
            % (op, recipe[NAME], recipe[SOURCE])
        )

    def remove_pattern(self, pattern_name):
        if pattern_name in self.runner_state[PATTERNS]:
            self.runner_state[PATTERNS].pop(pattern_name)
            self.remove_rules(deleted_pattern_name=pattern_name)
            runner_log(
                self.runner_state,
                'remove_pattern',
                'Removed pattern %s' % pattern_name
            )
        else:
            runner_log(
                self.runner_state,
                'remove_pattern',
                'Pattern %s was not present in the pattern list to be '
                'removed' % pattern_name
            )

    def remove_recipe(self, recipe_name):
        if recipe_name in self.runner_state[RECIPES]:
            self.runner_state[RECIPES].pop(recipe_name)
            self.remove_rules(deleted_recipe_name=recipe_name)
            runner_log(
                self.runner_state,
                'remove_recipe',
                'Removed recipe %s' % recipe_name
            )
        else:
            runner_log(
                self.runner_state,
                'remove_recipe',
                'Recipe %s was not present in the recipe list to be removed'
                % recipe_name
            )

    def create_new_rule(self, pattern_name, recipe_name, path):
        rule = {
            RULE_ID: generate_id(),
            RULE_PATTERN: pattern_name,
            RULE_RECIPE: recipe_name,
            RULE_PATH: path
        }
        _rules_lock.acquire()
        try:
            self.runner_state[RULES].append(rule)
            _rules_lock.release()
        except Exception as ex:
            _rules_lock.release()
            raise Exception(ex)

        runner_log(
            self.runner_state,
            'create_new_rule',
            'Created rule for path: %s with id %s.'
            % (path, rule[RULE_ID])
        )

        notebook_code = self.runner_state[RECIPES][rule[RULE_RECIPE]][RECIPE]

        pattern = self.runner_state[PATTERNS][rule[RULE_PATTERN]]

        yaml_dict = {}
        for var, val in pattern.variables.items():
            yaml_dict[var] = val
        for var, val in pattern.outputs.items():
            yaml_dict[var] = val
        yaml_dict[pattern.trigger_file] = path

        # print('setting up yaml dict')
        #
        # yaml_dict = {}
        # for var, val in pattern.variables.items():
        #     yaml_dict[var] = val
        # for var, val in pattern.outputs.items():
        #     yaml_dict[var] = val
        # yaml_dict[pattern.trigger_file] = src_path
        #
        # print('got yaml dict: %s' % yaml_dict)

        if self.runner_state[RETRO_ACTIVE]:
            testing_path = os.path.join(self.runner_state[VGRID], path)

            globbed = glob.glob(testing_path)

            for globble in globbed:
                yaml_dict[pattern.trigger_file] = globble

                local_path = globble[globble.find(os.path.sep)+1:]

                if not pattern.sweep:
                    schedule_job(
                        self.runner_state,
                        rule,
                        local_path,
                        notebook_code,
                        yaml_dict
                    )
                else:
                    for var, val in pattern.sweep.items():
                        values = get_parameter_sweep_values(val)
                        for value in values:
                            yaml_dict[var] = value
                            schedule_job(
                                self.runner_state,
                                rule,
                                local_path,
                                notebook_code,
                                yaml_dict
                            )

    def identify_rules(self, new_pattern=None, new_recipe=None):
        if new_pattern:
            if len(new_pattern.recipes) > 1:
                runner_log(
                    self.runner_state,
                    'identify_rules-pattern',
                    'Rule creation aborted. Currently only supports one '
                    'recipe per pattern.',
                )
            recipe_name = new_pattern.recipes[0]
            if recipe_name in self.runner_state[RECIPES]:
                for input_path in new_pattern.trigger_paths:
                    self.create_new_rule(
                        new_pattern.name,
                        recipe_name,
                        input_path
                    )

        if new_recipe:
            for name, pattern in self.runner_state[PATTERNS].items():
                if len(pattern.recipes) > 1:
                    runner_log(
                        self.runner_state,
                        'identify_rules-recipe',
                        'Rule creation avoided for %s. Currently only '
                        'supports one recipe per pattern.' % name,
                    )
                recipe_name = pattern.recipes[0]
                if recipe_name == new_recipe[NAME]:
                    for input_path in pattern.trigger_paths:
                        self.create_new_rule(
                            name,
                            recipe_name,
                            input_path
                        )

    def remove_rules(
            self, deleted_pattern_name=None, deleted_recipe_name=None):
        to_delete = []
        _rules_lock.acquire()
        try:
            for rule in self.runner_state[RULES]:
                if deleted_pattern_name:
                    if rule[RULE_PATTERN] == deleted_pattern_name:
                        to_delete.append(rule)
                if deleted_recipe_name:
                    if rule[RULE_RECIPE] == deleted_recipe_name:
                        to_delete.append(rule)
            for delete in to_delete:
                self.runner_state[RULES].remove(delete)
                runner_log(
                    self.runner_state,
                    'remove_rules',
                    'Removing rule: %s.' % delete
                )
            _rules_lock.release()
        except Exception as ex:
            _rules_lock.release()
            raise Exception(ex)


class LocalWorkflowMonitor(PatternMatchingEventHandler):
    """
    Event handler to schedule jobs according to file events.
    """

    def __init__(
            self, runner_state=None, patterns=None,
            ignore_patterns=None, ignore_directories=False,
            case_sensitive=False):
        """Constructor"""

        PatternMatchingEventHandler.__init__(
            self,
            patterns,
            ignore_patterns,
            ignore_directories,
            case_sensitive
        )
        self.runner_state = runner_state

        runner_log(
            runner_state,
            'LocalWorkflowRunner',
            'Starting new workflow runner'
        )

    def __handle_trigger(self, event, handle_path, rule):
        pid = multiprocessing.current_process().pid
        event_type = event.event_type
        src_path = event.src_path
        time_stamp = event.time_stamp

        runner_log(
            self.runner_state,
            '__handle_trigger',
            'Running threaded handler at (%s) to handle %s event at %s '
            'with rule %s'
            % (pid, event_type, handle_path, rule[RULE_ID]))

        # This will prevent some job spamming
        _recent_jobs_lock.acquire()
        try:
            if src_path in recent_jobs:
                if rule[RULE_ID] in recent_jobs[src_path]:
                    recent_timestamp = recent_jobs[src_path][rule[RULE_ID]]
                    difference = time_stamp - recent_timestamp

                    if difference <= 1:
                        recent_jobs[src_path][RULE_ID] = \
                            max(recent_timestamp, time_stamp)
                        runner_log(
                            self.runner_state,
                            '__handle_trigger',
                            'Skipping due to recent hit'
                        )
                        _recent_jobs_lock.release()
                        return
                else:
                    recent_jobs[src_path][rule[RULE_ID]] = time_stamp
            else:
                recent_jobs[src_path] = {
                    rule[RULE_ID]: time_stamp
                }
        except Exception as ex:
            _recent_jobs_lock.release()
            raise Exception(ex)
        _recent_jobs_lock.release()

        notebook_code = self.runner_state[RECIPES][rule[RULE_RECIPE]][RECIPE]

        pattern = self.runner_state[PATTERNS][rule[RULE_PATTERN]]

        runner_log(
            self.runner_state,
            'run_handler',
            'Starting new job for %s using rule %s' % (src_path, rule)
        )

        print('setting up yaml dict')

        yaml_dict = {}
        for var, val in pattern.variables.items():
            yaml_dict[var] = val
        for var, val in pattern.outputs.items():
            yaml_dict[var] = val
        yaml_dict[pattern.trigger_file] = src_path

        print('got yaml dict: %s' % yaml_dict)

        if not pattern.sweep:
            schedule_job(
                self.runner_state,
                rule,
                src_path,
                notebook_code,
                yaml_dict
            )
        else:
            for var, val in pattern.sweep.items():
                values = get_parameter_sweep_values(val)
                for value in values:
                    yaml_dict[var] = value
                    schedule_job(
                        self.runner_state,
                        rule,
                        src_path,
                        notebook_code,
                        yaml_dict
                    )

    def run_handler(self, event):
        src_path = event.src_path
        event_type = event.event_type

        handle_path = src_path.replace(self.runner_state[VGRID], '', 1)
        while handle_path.startswith(os.path.sep):
            handle_path = handle_path[1:]

        runner_log(
            self.runner_state,
            'run_handler',
            'Handling %s event at %s' % (event_type, handle_path)
        )

        for rule in self.runner_state[RULES]:
            target_path = rule[RULE_PATH]
            recursive_regexp = fnmatch.translate(target_path)
            direct_regexp = recursive_regexp.replace('.*', '[^/]*')
            recursive_hit = re.match(recursive_regexp, handle_path)
            direct_hit = re.match(direct_regexp, handle_path)

            if direct_hit or recursive_hit:
                waiting_for_thread_resources = True
                while waiting_for_thread_resources:
                    try:
                        worker = threading.Thread(
                            target=self.__handle_trigger,
                            args=(event, handle_path, rule))
                        worker.daemon = True
                        worker.start()
                        waiting_for_thread_resources = False
                    except threading.ThreadError as exc:
                        time.sleep(1)

    def handle_event(self, event):
        if event.is_directory:
            return

        if event.event_type not in ['created', 'modified']:
            return

        event.time_stamp = time.time()

        self.run_handler(event)

    def on_modified(self, event):
        """Handle modified files"""

        self.handle_event(event)

    def on_created(self, event):
        """Handle created files"""

        self.handle_event(event)

    def on_deleted(self, event):
        """Handle deleted files"""

        self.handle_event(event)

    def on_moved(self, event):
        """Handle moved files"""

        fake = make_fake_event(
            event.dest_path,
            'created',
            event.is_directory
        )
        self.handle_event(fake)



import copy
import json
import re
import os
import nbformat

from .constants import NOTEBOOK_EXTENSIONS, DESCENDANTS, WORKFLOW_INPUTS, \
    WORKFLOW_OUTPUTS, ANCESTORS, DEFAULT_JOB_NAME, CHAR_UPPERCASE, \
    CHAR_LOWERCASE, CHAR_NUMERIC, CHAR_LINES, PLACEHOLDER, TRIGGER_RECIPES, \
    NO_OUTPUT_SET_WARNING, NO_INPUT_FILE_SET_ERROR, NO_INPUT_PATH_SET_ERROR, \
    NO_NAME_SET_ERROR, NO_RECIPES_SET_ERROR, PLACEHOLDER_ERROR, \
    INVALID_INPUT_PATH_ERROR, SWEEP_START, SWEEP_STOP, SWEEP_JUMP, \
    MIG_TRIGGER_KEYWORDS, PERSISTENCE_ID, RECIPE_NAME, \
    PATTERN_NAME, INPUT_FILE, NAME, SWEEP, TRIGGER_PATHS, OUTPUT, RECIPES, \
    VARIABLES, RECIPE, SOURCE
from .validation import valid_string, is_valid_pattern_dict, \
    valid_file_path, valid_param_sweep, \
    check_input, is_valid_recipe_dict, valid_pattern_name, \
    valid_recipe_name
OUTPUT_MAGIC_CHARS = MIG_TRIGGER_KEYWORDS + ['*']


def is_valid_pattern_object(to_test, integrity=False):
    """
    Validates that the passed object is a meow Pattern

    :param: to_test: (dict) object, hopefully is a Pattern class object..

    :param: integrity: (boolean) toggle for if to perform integrity check on
    pattern object. Default is false.

    :return: (Tuple (bool, string) Returns a tuple where if the provided
    object is not a Pattern the first value will be False. Otherwise it will
    be True. If the first value is False then an explanatory error message is
    provided in the second value which will otherwise be an empty string.
    """

    if not to_test:
        return False, 'A workflow %s was not provided' % PATTERN_NAME

    if not isinstance(to_test, Pattern):
        return False, \
               'The workflow %s was incorrectly formatted' % PATTERN_NAME

    if integrity:
        return to_test.integrity_check()
    else:
        return True, ''


def check_patterns_dict(patterns, integrity=False):
    """
    Validates that the given object is a dictionary of valid patterns.

    :param: patterns: (dict) A dictionary of Pattern class objects.

    :param: integrity: (boolean) [optional] toggle for if to perform integrity
    check on pattern object. Default is false.

    :return: (Tuple (bool, string) Returns a tuple where if the provided
    object is not a dict of valid Pattern objects the first value will be
    False. Otherwise it will be True. If the first value is False then an
    explanatory error message is provided in the second value which will
    otherwise be an empty string.
    """
    if not isinstance(patterns, dict):
        return False, 'The provided %s(s) were not in a dict' % PATTERN_NAME
    for name, pattern in patterns.items():
        if not isinstance(pattern, Pattern):
            return False, \
                   'Object %s is a  %s, not the expected %s. '\
                   % (pattern, type(pattern), type(Pattern))

        if name != pattern.name:
            return False, \
                   '%s %s is not stored correctly as it does not share a ' \
                   'name with its key %s.' \
                   % (PATTERN_NAME, pattern.name, name)

        valid, feedback = is_valid_pattern_object(pattern, integrity=integrity)
        if not valid:
            return False, \
                   '%s %s was not valid. %s' \
                   % (PATTERN_NAME, pattern, feedback)
    return True, ''


def check_recipes_dict(recipes):
    """
    Validate that the given object is a dictionary of meow recipes.

    :param recipes: (dict) A dictionary of meow recipe dictionaries.

    :return: (Tuple (bool, string) Returns a tuple where if the provided
    object is not a dict of meow recipe dictionaries the first value will be
    False. Otherwise it will be True. If the first value is False then an
    explanatory error message is provided in the second value which will
    otherwise be an empty string.
    """
    if not isinstance(recipes, dict):
        return False, 'The provided %s(s) were not in a dict' % RECIPE_NAME
    else:
        for name, recipe in recipes.items():
            valid, feedback = is_valid_recipe_dict(recipe)
            if not valid:
                return False, \
                       '%s %s was not valid. %s' \
                       % (RECIPE_NAME, recipe, feedback)

            if name != recipe[NAME]:
                return False, \
                       '%s %s is not stored correctly as it does not share ' \
                       'a name with its key %s.' \
                       % (RECIPE_NAME, recipe[NAME], name)

    return True, ''


def parameter_sweep_entry(name, start, stop, increment):
    """
    Helper function to create parameter sweep entry

    :param name: (str) The name of the parameter sweep variable.

    :param start: (

    :return: (dict) A dictionary expressing a parameter sweep
    """
    check_input(name, str, 'name')

    return {
        SWEEP_START: start,
        SWEEP_STOP: stop,
        SWEEP_JUMP: increment
    }


def get_parameter_sweep_values(sweep):
    """

    :param sweep:
    :return:
    """
    valid_param_sweep(sweep, 'parameter_sweep')

    variables = []

    start = sweep[SWEEP_START]
    stop = sweep[SWEEP_STOP]
    increment = sweep[SWEEP_JUMP]

    par_val = start
    while par_val <= stop:
        variables.append(par_val)
        par_val += increment

    return variables


class Pattern:
    def __init__(self, parameters):
        """
        Constructor for new pattern object. Used within MEOW as a more user
        friendly and controllable way of creating and manipulating meow
        patterns, as opposed to a raw dict. Will raise a TypeError if
        incorrect parameters given.

        :param parameters: (string or dict) Takes a single input. This can be
        either a string, which is used as the patterns name when defining a
        new pattern, or can be a dict which already expresses a complete
        pattern. The dict option is used during the importing of data from
        the mig and should only be used by expert users.
        """
        # if given only a string use this as a name, it is the basis of a
        # completely new pattern
        if isinstance(parameters, str):
            valid_pattern_name(parameters)
            self.name = parameters
            self.trigger_file = None
            self.trigger_paths = []
            self.recipes = []
            self.outputs = {}
            self.variables = {}
            self.sweep = {}
        # if given dict we are importing from a stored pattern object
        elif isinstance(parameters, dict):
            valid, msg = is_valid_pattern_dict(parameters)

            if not valid:
                raise Exception("%s object is incorrectly formatted. %s"
                                % (PATTERN_NAME, msg))
            self.name = copy.deepcopy(parameters[NAME])

            if PERSISTENCE_ID in parameters:
                self.persistence_id = copy.deepcopy(parameters[PERSISTENCE_ID])

            self.trigger_paths = copy.deepcopy(parameters[TRIGGER_PATHS])
            self.recipes = []
            self.outputs = {}
            self.variables = {}
            self.sweep = {}

            self.trigger_file = copy.deepcopy(parameters[INPUT_FILE])

            # MiG has fun formatting recipe entries.
            if TRIGGER_RECIPES in parameters:
                for trigger_id, trigger in parameters[TRIGGER_RECIPES].items():
                    for recipe_key, recipe in trigger.items():
                        # A MiG recipe by this name has already been registered
                        if 'name' in recipe:
                            self.add_recipe(recipe['name'])
                        # If not already registered
                        else:
                            self.add_recipe(recipe_key)
            elif RECIPES in parameters:
                for recipe in parameters[RECIPES]:
                    self.add_recipe(recipe)

            if OUTPUT in parameters:
                for name, path in parameters[OUTPUT].items():
                    self.add_output(name, path)

            if VARIABLES in parameters:
                for name, value in parameters[VARIABLES].items():
                    self.add_variable(name, value)

            if SWEEP in parameters:
                for name, value in parameters[SWEEP].items():
                    self.add_param_sweep(name, value)
        else:
            raise TypeError(
                'Pattern requires either a str input as a name for a new '
                'pattern, or a dict defining a complete pattern. Was instead '
                'given a %s. ' % type(parameters)
            )

    def __str__(self):
        """
        Creates a string that expressed the current state of the Pattern
        object. Only displays parameters that should be directly editable by
        users.

        :return: (str) String expressing current Pattern object state.
        """
        string = 'Name: %s, ' \
                 'Input(s): %s, ' \
                 'Trigger(s): %s, ' \
                 'Output(s): %s, ' \
                 'Recipe(s): %s, ' \
                 'Variable(s): %s, ' \
                 'Sweeping(s): %s' \
                 % (self.name,
                    self.trigger_file,
                    self.trigger_paths,
                    self.outputs,
                    self.recipes,
                    self.variables,
                    self.sweep)
        return string

    def __eq__(self, other):
        """
        Tests if a given object is equal to this Pattern object.

        :param other: (Object) Object to test for equality.

        :return: (bool) True/False, with True denoting that the two Pattern
        objects are equal.
        """

        if not other:
            return False
        if not isinstance(other, Pattern):
            return False
        if self.name != other.name:
            return False
        if self.trigger_file != other.trigger_file:
            return False
        if self.trigger_paths != other.trigger_paths:
            return False
        if self.outputs != other.outputs:
            return False
        if self.recipes != other.recipes:
            return False
        if self.variables != other.variables:
            return False
        if self.sweep != other.sweep:
            return False
        count = 0
        try:
            self.persistence_id
            count += 1
        except AttributeError:
            pass
        try:
            other.persistence_id
            count += 1
        except AttributeError:
            pass
        if count == 1:
            return False
        if count == 2:
            if self.persistence_id != other.persistence_id:
                return False
        return True

    def display_dag_str(self):
        """
        Creates and display ready string, expressing the current Pattern
        state for a dag graph. Only displays the Pattern name, trigger file,
        trigger paths, outputs and recipes as any other characteristics are
        internal system parameters, or unnecessary for the dag.

        :return: (str) String expressing current Pattern object state.
        """
        string = 'Name: %s\n' \
                 'Input(s): %s\n' \
                 'Trigger(s): %s\n' \
                 'Output(s): %s\n' \
                 'Recipe(s): %s' \
                 % (self.name,
                    self.trigger_file,
                    self.trigger_paths,
                    self.outputs,
                    self.recipes)
        return string

    def integrity_check(self):
        """
        Performs some basic checks on the data within a pattern to check
        that all required fields have been filled out as it is currently very
        possible to create incomplete patterns.

        :return: (Tuple (bool, str)). Bool is a marker of if the pattern has
        passed the integrity check whilst str is a reason for a fail, or
        warnings of possible issues in the event of a pass.
        """
        warning = ''
        if self.name is None or not self.name:
            return False, NO_NAME_SET_ERROR
        if self.trigger_file is None or not self.trigger_file:
            return False, NO_INPUT_FILE_SET_ERROR
        if len(self.trigger_paths) == 0:
            return False, NO_INPUT_PATH_SET_ERROR
        for trigger_path in self.trigger_paths:
            if trigger_path is None or not trigger_path:
                return False, ("Error for input path '%s' " % trigger_path) \
                       + INVALID_INPUT_PATH_ERROR
        if len(self.outputs) == 0:
            warning += NO_OUTPUT_SET_WARNING
        if len(self.recipes) == 0:
            return False, NO_RECIPES_SET_ERROR

        if self.name == PLACEHOLDER \
                or self.trigger_file == PLACEHOLDER \
                or self.trigger_paths == PLACEHOLDER \
                or self.recipes == PLACEHOLDER \
                or PLACEHOLDER in self.trigger_paths \
                or PLACEHOLDER in self.outputs.keys() \
                or PLACEHOLDER in self.outputs.values() \
                or PLACEHOLDER in self.recipes \
                or PLACEHOLDER in self.variables.keys() \
                or PLACEHOLDER in self.variables.values():
            return False, PLACEHOLDER_ERROR

        return True, warning

    def add_single_input(self, input_file, regex_path, output_path=None):
        """
        Defines a single input for a pattern. That is, a path that when a
        single file is either created or modified and matches said path, the
        recipe will be triggered taking that file as input. Raises
        AttributeError if an input is already defined.

        :param input_file: (str) 'input_file' is the variable name used to
        refer to the triggering file within the recipe.

        :param regex_path: (str) 'regex_path' is path against which any file
        creation or modification events are compared. The path should be given
        relative to the vgrid home directory so the path 'dir/file.txt' would
        match the file 'file.txt' within the directory 'dir' that is contained
        within the vgrid folder itself. Paths can, and should be expressed
        using regular expressions in order to match multiple files and so
        create a dynamic workflow.
        For example:

        initial_data/.*\\.hdf5

        This would match any hdf5 files contained within the directory initial
        data. Note that as string expression of regular expression normal use
        of escape characters and the like apply.

        :param output_path: (str)[optional] 'output_path' may also be defined.
        If done so, this will be the path where the input file will be copied
        after job completion. This is useful if some data processing is taking
        place on the input and you require that data, as the triggering file
        itself will remain unchanged. Note that this is a regular string path
        and can be hard coded such as 'dir/file.txt' which will always write
        to the same location. Alternatively the '*' character can be used to
        take the name of the triggering file.
        For example:

        'output_dir/*.hdf5'

        When triggered by a file called 'filename.txt' this would copy the
        output to output_dir/filename.hdf5. Default is None.

        :return: No return.
        """

        check_input(input_file, str, 'input_file')
        check_input(regex_path, str, 'regex_path')
        check_input(output_path, str, 'output_path', or_none=True)

        if len(self.trigger_paths) == 0:
            self.trigger_file = input_file
            self.trigger_paths = [regex_path]
            if output_path:
                self.add_output(input_file, output_path)
        else:
            raise AttributeError(
                'Could not create single input %s, as input already defined'
                % input_file
            )

    def add_output(self, output_name, output_location):
        """
        Adds output to the pattern. That is, defines some file that is copied
        as output from the job processing. If any file is created by the
        recipe at runtime is should be added as output using this method or
        else it may be lost.

        :param output_name: (str) variable name used within the recipe to
        refer to the output file.

        :param output_location: (str) path where the file will be copied on
        job completion. Note that this is a string path and can be hard coded
        such as 'dir/file.txt' which will always write to the same location.
        Alternatively the '*' character can be used to take the name of the
        triggering file.
        For example:

        'output_dir/*.hdf5'

        When triggered by a file called 'filename.txt' this would copy the
        output to output_dir/filename.hdf5.

        :return: No return.
        """
        check_input(output_name, str, 'output_name')
        check_input(output_location, str, 'output_location')

        if output_name not in self.outputs.keys():
            self.outputs[output_name] = output_location
        else:
            raise Exception('Could not create output %s as already defined'
                            % output_name)

    def return_notebook(self, output_location):
        """
        Adds the notebook used to run the job as output.

        :param output_location: (str) The path where the file will be copied
        on job completion. Note that this is a string path and can be hard
        coded such as 'dir/file.txt' which will always write to the same
        location. Alternatively the '*' character can be used to take the name
        of the triggering file.
        For example:

        'output_dir/*.hdf5'

        When triggered by a file called 'filename.txt' this would copy the
        output to output_dir/filename.hdf5.

        :return: No return.
        """
        check_input(output_location, str, 'output_location')
        self.add_output(DEFAULT_JOB_NAME, output_location)

    def add_recipe(self, recipe):
        """
        Adds a recipe to the pattern. This is the code that runs as part of a
        workflow job, and is triggered by the patterns specified trigger.

        :param recipe: (str) Name of recipe to be added.

        :return: No return.
        """
        check_input(recipe, str, 'recipe')
        self.recipes.append(recipe)

    def add_variable(self, variable_name, variable_value):
        """
        Adds a variable to the pattern, which will be passed to the recipe
        notebook using papermill as parameters. Raises ValueError if variable
        name is already in use.

        :param variable_name: (str) name of the variable.

        :param variable_value: (any) any valid base python variable value.

        :return: No return.
        """
        valid_string(variable_name,
                     'variable name',
                     CHAR_UPPERCASE
                     + CHAR_LOWERCASE
                     + CHAR_NUMERIC
                     + CHAR_LINES)
        if variable_name not in self.variables.keys():
            self.variables[variable_name] = variable_value
        else:
            raise ValueError(
                'Could not create variable %s as a variable with this '
                'name is already defined. ' % variable_name
            )

    def add_param_sweep(self, name, sweep_dict):
        """
        Adds a parameter sweeping variable to the Pattern, which will be
        passed to the Recipe notebook using papermill as parameters. Raises
        ValueError if parameter sweeping variable name is already in use.

        :param name: (str) name of the variable.

        :param sweep_dict: (dict) dictionary defining the parameter sweep.
        Should be of the form:
        {
            'start': (int) or (float),
            'stop': (int) or (float),
            'jump': (int) or (float)
        }

        :return: No return.
        """
        valid_string(name,
                     'name',
                     CHAR_UPPERCASE
                     + CHAR_LOWERCASE
                     + CHAR_NUMERIC
                     + CHAR_LINES)
        valid_param_sweep(sweep_dict, name)

        if name not in self.sweep.keys():
            self.sweep[name] = sweep_dict
        else:
            raise ValueError(
                'Could not create parameter sweeping variable %s as a '
                'parameter sweeping variable with this name is already '
                'defined. ' % name
            )

    def to_display_dict(self):
        """
        Creates a dictionary of the current pattern state to be displayed as
        part of the WorkflowWidget visualisation.

        :return: (dict) Dictionary expressing current state of the Pattern
        object.
        """
        display_dict = {
            NAME: self.name,
            INPUT_FILE: self.trigger_file,
            TRIGGER_PATHS: self.trigger_paths,
            RECIPES: self.recipes,
            SWEEP: self.sweep,
            OUTPUT: {},
            VARIABLES: {}
        }

        for key, value in self.variables.items():
            if key not in self.outputs \
                    and key != self.trigger_file:
                display_dict[VARIABLES][key] = value

        for key, value in self.outputs.items():
            if key != self.trigger_file and key != DEFAULT_JOB_NAME:
                display_dict[OUTPUT][key] = value

        return display_dict


def register_recipe(source, name=None):
    """
    Registers a new Recipe dict.

    :param source: (str) The source notebook of the Recipe.

    :param name: (str)[optional] The name the Recipe is to be registered under.
    If not provided then the source filename will be used instead.

    :return: (dict) A dict object expressing the Recipe notebook at the given
    source
    """
    valid_file_path(source, 'source', extensions=NOTEBOOK_EXTENSIONS)
    check_input(name, str, 'name', or_none=True)

    if not name:
        if os.path.sep in source:
            name = source[source.rfind('/') + 1:source.index('.')]
        else:
            name = source[:source.index('.')]
    if not os.path.isfile(source):
        raise FileNotFoundError("Source %s was not found. " % source)

    with open(source, "r") as read_file:
        notebook = json.load(read_file)
        return create_recipe_dict(notebook, name, source)


def create_recipe_dict(notebook, name, source):
    """
    Creates a recipe dictionary from the given parameters.

    :param notebook: (dict) notebook source code.

    :param name: (str) name of recipe.

    :param source: (str) original notebook source code was extracted from.

    :return: (dict) Meow recipe dictionary. Format is {'name': str,
    'source': str, 'recipe': dict}
    """

    valid_recipe_name(name)
    valid_file_path(
        source,
       'recipe source'
    )
    nbformat.validate(notebook)

    recipe = {
        NAME: name,
        SOURCE: source,
        RECIPE: notebook
    }
    return recipe


def build_workflow_object(patterns, vgrid=None):
    """
    Builds the emergent workflow from defined patterns.

    :param patterns: (dict) A dictionary of valid Pattern objects.

    :param vgrid: (str) [Optional] The name of the vgrid this workflow is
    operating in

    :return: (Tuple (bool, string or dict) Returns a tuple with the first value
    being a boolean, with True showing that a workflow was built without
    errors, and False showing that a problem was encountered. If False the
    second value is an explanatory error message. If True the second value is
    the produced workflow, which may be empty if no workflow can be built from
    the pattern definitions. The workflow is a dict of nodes with each node
    being a dict with Format {'descendants': dict, 'ancestors': dict,
    'workflow inputs': dict, 'workflow outputs': dict}.
    """

    valid, msg = check_patterns_dict(patterns)
    if not valid:
        return False, msg

    workflow = {}
    # create all required nodes
    for pattern in patterns.values():
        input_paths = {}
        output_paths = {}
        input_paths[pattern.trigger_file] = pattern.trigger_paths
        for file, path in pattern.outputs.items():
            output_paths[file] = path
        workflow[pattern.name] = {
            DESCENDANTS: {},
            ANCESTORS: {},
            WORKFLOW_INPUTS: input_paths,
            WORKFLOW_OUTPUTS: output_paths
        }

    # populate nodes with ancestors and descendants
    for pattern in patterns.values():
        input_regex_list = pattern.trigger_paths
        for other in patterns.values():
            other_output_dict = other.outputs
            for input_regex in input_regex_list:
                # To match inputs to outputs we need to add on the vgrid path.
                # We might be able to guess this if self.vgrid is defined,
                # otherwise just use the placeholder term.
                if vgrid:
                    full_input_regex = os.path.join(vgrid, input_regex)
                else:
                    full_input_regex = os.path.join('{VGRID}', input_regex)

                for key, value in other_output_dict.items():
                    filename = value
                    if os.path.sep in filename:
                        filename = filename[filename.rfind(os.path.sep)+1:]
                    match_dict = {
                        'output_pattern': other.name,
                        'output_file': key,
                        'value': value,
                        'filename': filename
                    }
                    if re.match(full_input_regex, value):
                        workflow[other.name][DESCENDANTS][pattern.name] = \
                            match_dict
                        workflow[pattern.name][ANCESTORS][other.name] = \
                            match_dict
                        if pattern.trigger_file in \
                                workflow[pattern.name][WORKFLOW_INPUTS]:
                            workflow[pattern.name][WORKFLOW_INPUTS].pop(
                                pattern.trigger_file
                            )
                    check = [mc for mc in OUTPUT_MAGIC_CHARS if mc in value]
                    if check:
                        magic_value = value
                        for magic_char in check:
                            magic_value = magic_value.replace(magic_char, '.*')
                        if re.match(magic_value, full_input_regex):
                            workflow[other.name][DESCENDANTS][pattern.name] = \
                                match_dict
                            workflow[pattern.name][ANCESTORS][other.name] = \
                                match_dict
                            if pattern.trigger_file in \
                                    workflow[pattern.name][WORKFLOW_INPUTS]:
                                workflow[pattern.name][WORKFLOW_INPUTS].pop(
                                    pattern.trigger_file
                                )
    return True, workflow


def pattern_has_recipes(pattern, recipes):
    """
    Checks that a pattern has all required recipes in the workflow for it
    to be triggerable

    :param pattern: (Pattern) A pattern object.

    :param recipes: (dict) A dictionary of recipes.

    :return: (Tuple (bool, string) Returns a tuple with the first value
    being a boolean, with True showing that all pattern recipes are already
    registered, and False showing that a problem was encountered. If False the
    second value is an explanatory error message. If True the second value is
    an empty string.
    """

    valid, feedback = is_valid_pattern_object(pattern)

    if not valid:
        return False, "Pattern %s is not valid. %s" % (pattern, feedback)

    if not recipes:
        return False, 'Recipes were not provided.'

    if not isinstance(recipes, dict):
        return False, \
               'Recipes was not formatted correctly. Expected dict but got ' \
               '%s. ' % type(recipes)

    for recipe in recipes.values():
        if not isinstance(recipe, dict):
            return False, \
                   'Recipe %s was incorrectly formatted. Expected %s but ' \
                   'got %s. ' % (recipe, dict, type(recipe))
        valid, feedback = is_valid_recipe_dict(recipe)
        if not valid:
            return False, "Recipe %s is not valid. %s" % (recipe, feedback)

    for recipe in pattern.recipes:
        if recipe not in recipes:
            return False, 'Recipe %s was not present in recipes. '
    return True, ''




from .constants import CWL_NAME, CWL_CWL_VERSION, CWL_CLASS, CWL_BASE_COMMAND,\
    CWL_INPUTS, CWL_OUTPUTS, CWL_STEPS, CWL_HINTS, CWL_REQUIREMENTS, \
    CWL_ARGUMENTS, CWL_STDOUT, CWL_WORKFLOW_IN, CWL_WORKFLOW_OUT, \
    CWL_YAML_PATH, CWL_OUTPUT_BINDING, CWL_OUTPUT_TYPE, CWL_OUTPUT_GLOB, \
    CWL_VARIABLES, PLACEHOLDER, WORKFLOW_NAME, STEP_NAME, VARIABLES_NAME, \
    WORKFLOWS, STEPS, SETTINGS, CWL_CLASS_COMMAND_LINE_TOOL, \
    CWL_CLASS_WORKFLOW, CWL_WORKFLOW_RUN
from .validation import check_input, is_valid_workflow_dict, \
    is_valid_step_dict, is_valid_setting_dict


def make_step_dict(name, base_command):
    """
    Creates a new dict defining a base CWL step.

    :param name: (str) Name of CWL step.

    :param base_command: (str) Base command for CWL step.

    :return: (dict) Dictionary defining CWL step. Format is:
    {
        'name': str,
        'cwlVersion': str,
        'class': str,
        'baseCommand': str,
        'stdout': str,
        'inputs': dict,
        'outputs': dict,
        'arguments': list,
        'requirements': dict,
        'hints': dict
    }
    """
    check_input(name, str, 'name')
    check_input(base_command, str, 'baseCommand')

    return {
        CWL_NAME: name,
        CWL_CWL_VERSION: 'v1.0',
        CWL_CLASS: CWL_CLASS_COMMAND_LINE_TOOL,
        CWL_BASE_COMMAND: base_command,
        CWL_STDOUT: '',
        CWL_INPUTS: {},
        CWL_OUTPUTS: {},
        CWL_ARGUMENTS: [],
        CWL_REQUIREMENTS: {},
        CWL_HINTS: {},
    }


def make_workflow_dict(name):
    """
    Creates a new dict defining a base CWL workflow.

    :param name: (str or dict) Name of CWL workflow.

    :return: (dict) Dictionary defining CWL workflow. Format is:
    {
        'name': str,
        'cwlVersion': str,
        'class': str,
        'inputs': dict,
        'outputs': dict,
        'requirements': dict,
        'steps': dict
    }
    """

    check_input(name, str, 'name')

    return {
        CWL_NAME: name,
        CWL_CWL_VERSION: 'v1.0',
        CWL_CLASS: CWL_CLASS_WORKFLOW,
        CWL_INPUTS: {},
        CWL_OUTPUTS: {},
        CWL_STEPS: {},
        CWL_REQUIREMENTS: {}
    }


def make_settings_dict(name, yaml):
    """
    Creates a new dict defining CWL input variables to be saved as a yaml file.

    :param name: (str) Name of settings dict. This is used to match the dict
    to a workflow or step file.

    :param yaml: (dict) Variables to be saved.

    :return: (dict) Dictionary input arguments. Format is:
    {
        'name': str,
        'arguments': dict
    }
    """

    check_input(name, str, 'name')
    check_input(yaml, dict, 'settings', or_none=True)

    return {
        CWL_NAME: name,
        CWL_VARIABLES: yaml
    }


def get_linked_workflow(workflow, steps, settings):
    """
    Determines static workflow using CWL definitions. This will attempt to get
    all steps defined in the 'workflow' and construct a dict of nodes to be
    displayed.

    :param workflow: (dict) A CWL workflow dictionary.

    :param steps: (dict) A dict of CWL step dictionaries.

    :param settings: (dict) A CWL arguments dictionary.

    :return: Tuple (dict) Returns a dict of workflow nodes. Format is:
    {
        'inputs': dict,
        'outputs': dict,
        'ancestors': dict
    }
    """

    workflow_nodes = {}

    for step_title, step in workflow[CWL_STEPS].items():
        step_name = get_step_name_from_title(step_title, workflow)
        workflow_nodes[step_name] = {
            'inputs': {},
            'outputs': {},
            'ancestors': {},
        }

        for input_key, input_value in step[CWL_WORKFLOW_IN].items():
            if '/' not in input_value:
                if workflow[CWL_INPUTS][input_value] == 'File':
                    workflow_nodes[step_name]['inputs'][input_key] = \
                        settings[input_value][CWL_YAML_PATH]
            else:
                ancestor = input_value[:input_value.index('/')]
                workflow_nodes[step_name]['ancestors'][ancestor] = input_value

        # TODO clean this up. I hate it and its ugly
        outputs_string = step[CWL_WORKFLOW_OUT]

        outputs_list = []
        if isinstance(outputs_string, list):
            outputs_list = outputs_string
        elif isinstance(outputs_string, str):
            if outputs_string.startswith('[') and outputs_string.endswith(']'):
                outputs_string = outputs_string[1:-1]

                outputs_list = outputs_string.split(',')
                for i in range(len(outputs_list)):
                    outputs_list[i] = outputs_list[i].replace(' ', '')

        for output in outputs_list:
            full_output = '%s/%s' % (step_name, output)
            if step_name not in steps:
                break
            if steps[step_name][CWL_OUTPUTS][output][CWL_OUTPUT_TYPE] \
                    == 'File':
                output_value = \
                    steps[step_name][CWL_OUTPUTS][output][CWL_OUTPUT_BINDING]
                if isinstance(output_value, dict):
                    glob = output_value[CWL_OUTPUT_GLOB]
                    if glob.startswith('$(inputs'):
                        glob = glob[glob.index('s')+1:glob.index(')')]

                        if glob.startswith('.'):
                            glob = glob[1:]
                        if glob.startswith('[') and glob.endswith(']'):
                            glob = glob[1:-1]
                        if glob.startswith('\'') and glob.endswith('\''):
                            glob = glob[1:-1]
                        if glob.startswith('"') and glob.endswith('"'):
                            glob = glob[1:-1]
                        if glob in step[CWL_WORKFLOW_IN]:
                            output_key = step[CWL_WORKFLOW_IN][glob]
                            output_value = settings[output_key]
                    else:
                        break
                workflow_nodes[step_name]['outputs'][full_output] = \
                    output_value

    return workflow_nodes


def get_step_name_from_title(title, workflow):
    """
    Gets the specific name of a step according to the longer auto-generated
    but definately unique title.

    :param title: (str)

    :param workflow: (dict) A CWL workflow dict.

    :return: (str) Name derived from title, according to workflow definition.
    """
    name = workflow[CWL_STEPS][title][CWL_WORKFLOW_RUN]
    if '.' in name:
        name = name[:name.index('.')]
    return name


def check_workflow_is_valid(workflow_name, cwl):
    """
    Checks that a given workflow is valid, with a corresponding arguments
    definiton and that no placeholder values are present.

    :param workflow_name: (str)

    :param cwl: (dict) A dict of CWL defitions, including workflows, steps and
    arguments.

    :return: (Tuple (bool, str) If workflow is valid returns a tuple of first
    value True and second value an empty string. Else, returns a tuple of
    first value False, with the second value being an explanatory error
    message.
    """

    if workflow_name not in cwl[WORKFLOWS]:
        msg = "%s \'%s\' does not exist within the current CWL definitions. " \
              % (WORKFLOW_NAME, workflow_name)
        return False, msg
    workflow = cwl[WORKFLOWS][workflow_name]

    if workflow_name not in cwl[SETTINGS]:
        msg = "%s \'%s\' does not have a corresponding YAML %s definition. " \
              "This should share the name of the %s, and in this case " \
              "should be \'%s\'" \
              % (WORKFLOW_NAME, workflow_name, VARIABLES_NAME, WORKFLOW_NAME,
                 workflow_name)
        return False, msg
    settings = cwl[SETTINGS][workflow_name]

    for input_key, input_type in workflow[CWL_INPUTS].items():
        if input_key not in settings[CWL_VARIABLES]:
            msg = '%s \'%s\' expects input \'%s\', but it is not present in ' \
                  'yaml dict \'%s\'. ' \
                  % (WORKFLOW_NAME, workflow[CWL_NAME], input_key,
                     settings[CWL_NAME])
            return False, msg
        if settings[CWL_VARIABLES][input_key] == PLACEHOLDER:
            msg = '%s \'%s\' contains an entry for \'%s\' whose value is ' \
                  'the placeholder value \'%s\'. Please update this to an ' \
                  'actual value before proceeding. ' \
                  % (VARIABLES_NAME, settings[CWL_NAME], input_key,
                     PLACEHOLDER)
            return False, msg
        if isinstance(settings[CWL_VARIABLES][input_key], dict):
            for key, value in settings[CWL_VARIABLES][input_key].items():
                if value == PLACEHOLDER:
                    msg = '%s \'%s\' contains a dict for \'%s\' which ' \
                          'contains the key \'%s\', whose value is the ' \
                          'placeholder value \'%s\'. Please update this to ' \
                          'an actual value before proceeding. ' \
                          % (VARIABLES_NAME, settings[CWL_NAME], input_key,
                             key, PLACEHOLDER)
                    return False, msg
    return True, ''


def check_step_is_valid(step_name, cwl):
    """
    Checks that a given step is valid.

    :param step_name: (str) Name of step to check.

    :param cwl: (dict) Dictionary of CWL steps dictionaries.

    :return: (Tuple (bool, str) If step is valid returns a tuple of first
    value True and second value an empty string. Else, returns a tuple of
    first value False, with the second value being an explanatory error
    message.
    """

    if step_name not in cwl[STEPS]:
        msg = "%s \'%s\' does not exist within the current CWL definitions. " \
              % (STEP_NAME, step_name)
        return False, msg

    return True, ''


def get_glob_value(glob, step_name, workflow_cwl, settings_cwl):
    """
    Attempts to get the value associated with a glob gathering operation given
    the currently defined workflow, steps and arguments.

    :param glob: (str) The glob definition string.

    :param step_name: (str) The name of the step the glob variable is derived
    from.

    :param workflow_cwl: (dict) A CWL workflow dictionary.

    :param settings_cwl: (dict) A CWL arguments dictionary.

    :return: (Tuple(bool, string)) If 'glob' input was not formatted
    correctly or could not be found then a tuple will be returned with the
    first value being False, and the second being an appropriate error
    message. If no problems are encountered then a tuple is returned with the
    first value being True and the appropriate argument in the second value.
    """

    if '$' in glob:
        glob = glob[glob.index('(') + 1:glob.index(')')]
        inputs = glob.split('.')
        if inputs[0] != CWL_INPUTS:
            msg = "Unexpected path. %s did not equal expected path " \
                  "%s" % (inputs[0], CWL_INPUTS)
            return False, msg

        settings_key = ''
        workflow_steps = list(workflow_cwl[CWL_STEPS].keys())
        for step in workflow_steps:
            if step == step_name:
                settings_key = \
                    workflow_cwl[CWL_STEPS][step][CWL_WORKFLOW_IN][inputs[1]]

        if not settings_key:
            msg = "Could not find step %s in %s" \
                  % (step_name, list(workflow_cwl[CWL_STEPS]))
            return False, msg
        setting = settings_cwl[settings_key]

        return True, setting
    else:
        return True, glob


def get_glob_entry_keys(glob, step_name, workflow_cwl):
    """
    Attempts to get a key/value pair from a glob definition. This is the pair
    of lookups within the appropriate arguments dictionary when the CWL is
    imported from MEOW definitions.

    :param glob: (str) The glob definition string.

    :param step_name: (str) The name of the step the glob variable is derived
    from.

    :param workflow_cwl: (dict) A CWL workflow dictionary.

    :return: (Tuple(bool, string or Tuple(string, string)) If 'glob'
    input was not formatted correctly or could not be found then a tuple will
    be returned with the first value being False, and the second being an
    appropriate error message. If no problems are encountered then a tuple is
    returned with the first value being True and a tuple of the key/value
    pairs being the the second value.
    """
    if '$' in glob:
        try:
            glob = glob[glob.index('(') + 1:glob.index(')')]
            inputs = glob.split('.')
            if inputs[0] != CWL_INPUTS:
                msg = "Unexpected path. %s did not equal expected path " \
                      "%s" % (inputs[0], CWL_INPUTS)
                return False, msg
            settings_key = \
                workflow_cwl[CWL_STEPS][step_name][CWL_WORKFLOW_IN][inputs[1]]

            if settings_key.endswith('_key'):
                entry = settings_key[:settings_key.rfind('_')]
                value_key = "%s_value" % entry
                return True, (settings_key, value_key)

            elif settings_key.endswith('_value'):
                entry = settings_key[:settings_key.rfind('_')]
                key_key = "%s_key" % entry
                return True, (key_key, settings_key)

        except Exception as exception:
            return False, str(exception)
    msg = 'Could not identify matching key pairs for glob %s. ' % glob
    return False, msg


def get_output_lookup(target_step_key, target_value, workflow, steps):
    """
    Retrieves the output value for a given step with a given value.

    :param target_step_key: (str) Name of the CWL step.

    :param target_value: (str) Name of the CWL step output.

    :param workflow: (dict) CWL Workflow dict containing the step.

    :param steps: (dict) CWL steps dict.

    :return: (Tuple(bool, string)) If output was not found then a tuple will
    be returned with the first value being False, and the second being an
    appropriate error message. If no problems are encountered then a tuple is
    returned with the first value being True and the output argument in the
    second value.
    """
    if target_step_key in workflow[CWL_STEPS]:
        target_step_key = get_step_name_from_title(
            target_step_key,
            workflow
        )
    if target_step_key not in steps:
        msg = "%s isn't in steps" % target_step_key
        return False, msg
    target_step = steps[target_step_key]
    if target_value not in target_step[CWL_OUTPUTS]:
        msg = '%s not in outputs' % target_value
        return False, msg
    target_output = target_step[CWL_OUTPUTS][target_value]

    if CWL_OUTPUT_TYPE not in target_output \
            or target_output[CWL_OUTPUT_TYPE] != 'File' \
            or CWL_OUTPUT_BINDING not in target_output \
            or not isinstance(target_output[CWL_OUTPUT_BINDING], dict) \
            or 'glob' not in \
            target_output[CWL_OUTPUT_BINDING]:
        msg = 'Output type is not a correctly formatted file entry'
        return False, msg
    glob = target_output[CWL_OUTPUT_BINDING]['glob']
    return True, glob


def check_workflows_dict(workflows):
    """
    Validate that the given object is a dictionary of cwl workflows.

    :param workflows: (dict) A dictionary of cwl workflow dictionaries.

    :return: (Tuple (bool, string) Returns a tuple where if the provided
    object is not a dict of cwl workflows dictionaries the first value will be
    False. Otherwise it will be True. If the first value is False then an
    explanatory error message is provided in the second value which will
    otherwise be an empty string.
    """
    if not isinstance(workflows, dict):
        return False, 'The provided %s(s) were not in a dict' % WORKFLOW_NAME
    else:
        for name, workflow in workflows.items():
            valid, feedback = is_valid_workflow_dict(workflow)
            if not valid:
                return False, \
                       '%s %s was not valid. %s' \
                       % (WORKFLOW_NAME, workflow, feedback)
            if name != workflow[CWL_NAME]:
                return False, \
                       '%s %s is not stored correctly as it does not share ' \
                       'a name with its key %s.' \
                       % (WORKFLOW_NAME, workflow[CWL_NAME], name)
    return True, ''


def check_steps_dict(steps):
    """
    Validate that the given object is a dictionary of cwl steps.

    :param steps: (dict) A dictionary of cwl step dictionaries.

    :return: (Tuple (bool, string) Returns a tuple where if the provided
    object is not a dict of cwl steps dictionaries the first value will be
    False. Otherwise it will be True. If the first value is False then an
    explanatory error message is provided in the second value which will
    otherwise be an empty string.
    """
    if not isinstance(steps, dict):
        return False, 'The provided %s(s) were not in a dict' % STEP_NAME
    else:
        for name, step in steps.items():
            valid, feedback = is_valid_step_dict(step)
            if not valid:
                return False, \
                       '%s %s was not valid. %s' \
                       % (STEP_NAME, step, feedback)
            if name != step[CWL_NAME]:
                return False, \
                       '%s %s is not stored correctly as it does not share ' \
                       'a name with its key %s.' \
                       % (WORKFLOW_NAME, step[CWL_NAME], name)
    return True, ''


def check_settings_dict(settings):
    """
    Validate that the given object is a dictionary of cwl settings.

    :param settings: (dict) A dictionary of cwl argument dictionaries.

    :return: (Tuple (bool, string) Returns a tuple where if the provided
    object is not a dict of cwl arguments dictionaries the first value will be
    False. Otherwise it will be True. If the first value is False then an
    explanatory error message is provided in the second value which will
    otherwise be an empty string.
    """
    if not isinstance(settings, dict):
        return False, 'The provided %s(s) were not in a dict' % VARIABLES_NAME
    else:
        for name, setting in settings.items():
            valid, feedback = is_valid_setting_dict(setting)
            if not valid:
                return False, \
                       '%s %s was not valid. %s' \
                       % (VARIABLES_NAME, setting, feedback)
            if name != setting[CWL_NAME]:
                return False, \
                       '%s %s is not stored correctly as it does not share ' \
                       'a name with its key %s.' \
                       % (WORKFLOW_NAME, setting[CWL_NAME], name)
    return True, ''




import ipywidgets as widgets
import time

from IPython.display import display

from .validation import check_input
from .constants import VGRID_READ, VGRID_QUEUE_OBJECT_TYPE,\
    VGRID_CREATE, VGRID_UPDATE, VGRID_JOB_OBJECT_TYPE, \
    OBJECT_TYPE
from .logging import create_monitor_logfile, write_to_log
from .mig import vgrid_job_json_call


MRSL_VGRID = 'VGRID'
MRSL_JOB_ID = 'JOB_ID'
MRSL_JOB_STATUS = 'STATUS'
MRSL_JOB_RECEIVED_TIME = 'RECEIVED_TIMESTAMP'
MRSL_JOB_VGRID = 'VGRID'
MRSL_JOB_EXECUTE = 'EXECUTE'
MRSL_JOB_EXECUTABLES = 'EXECUTABLES'
MRSL_JOB_INPUTFILES = 'INPUTFILES'
MRSL_JOB_OUTPUTFILES = 'OUTPUTFILES'
MRSL_JOB_NAME = 'JOBNAME'
MRSL_JOB_RETRIES = 'RETRIES'
MRSL_JOB_RUNTIMEENVIROMENT = 'RUNTIMEENVIRONMENT'
MRSL_JOB_ENVIROMENT = 'ENVIRONMENT'
MRSL_JOB_VERIFYFILES = 'VERIFYFILES'
MRSL_JOB_DISK = 'DISK'
MRSL_JOB_JOBTYPE = 'JOBTYPE'
MRSL_JOB_RESOURCE = 'RESOURCE'
MRSL_JOB_CPUTIME = 'CPUTIME'
MRSL_JOB_PLATFORM = 'PLATFORM'
MRSL_JOB_USERCERT = 'USER_CERT'
MRSL_JOB_MAXFILL = 'MAXFILL'
MRSL_JOB_MOUNT = 'MOUNT'
MRSL_JOB_MAXPRICE = 'MAXPRICE'
MRSL_JOB_PROJECT = 'PROJECT'
MRSL_JOB_SANDBOX = 'SANDBOX'
MRSL_JOB_CPUCOUNT = 'CPUCOUNT'
MRSL_JOB_NOTIFY = 'NOTIFY'
MRSL_JOB_NODECOUNT = 'NODECOUNT'
MRSL_JOB_ARCHITECTURE = 'ARCHITECTURE'
MRSL_JOB_MEMORY = 'MEMORY'
MRSL_JOB_QUEUED_TIME = 'QUEUED_TIMESTAMP'
MRSL_JOB_CANCELED_TIME = 'CANCELED_TIMESTAMP'

JOB_QUEUE_KEYS = {
    MRSL_JOB_ID: 'Job ID',
    MRSL_JOB_STATUS: 'Status',
    MRSL_JOB_RECEIVED_TIME: 'Created at'
}

JOB_CORE_DISPLAY_KEYS = {
    MRSL_JOB_ID: 'Job ID',
    MRSL_JOB_STATUS: 'Status',
    MRSL_JOB_RECEIVED_TIME: 'Created at',
    MRSL_JOB_VGRID: 'Vgrid',
    MRSL_JOB_EXECUTE: 'Execute',
    MRSL_JOB_INPUTFILES: 'Input files',
    MRSL_JOB_EXECUTABLES: 'Executable files',
    MRSL_JOB_OUTPUTFILES: 'Output files',
}

JOB_KEYS = [
    MRSL_JOB_ID,
    MRSL_JOB_STATUS,
    MRSL_JOB_RECEIVED_TIME,
    MRSL_JOB_VGRID,
    MRSL_JOB_EXECUTE,
    MRSL_JOB_EXECUTABLES,
    MRSL_JOB_INPUTFILES,
    MRSL_JOB_OUTPUTFILES,
    MRSL_JOB_NAME,
    MRSL_JOB_RETRIES,
    MRSL_JOB_RUNTIMEENVIROMENT,
    MRSL_JOB_ENVIROMENT,
    MRSL_JOB_VERIFYFILES,
    MRSL_JOB_DISK,
    MRSL_JOB_JOBTYPE,
    MRSL_JOB_RESOURCE,
    MRSL_JOB_CPUTIME,
    MRSL_JOB_PLATFORM,
    MRSL_JOB_USERCERT,
    MRSL_JOB_MAXFILL,
    MRSL_JOB_MOUNT,
    MRSL_JOB_MAXPRICE,
    MRSL_JOB_PROJECT,
    MRSL_JOB_SANDBOX,
    MRSL_JOB_CPUCOUNT,
    MRSL_JOB_NOTIFY,
    MRSL_JOB_NODECOUNT,
    MRSL_JOB_ARCHITECTURE,
    MRSL_JOB_MEMORY,
    MRSL_JOB_QUEUED_TIME,
    MRSL_JOB_CANCELED_TIME
]

SUBMISSION_JOB_KEYS = [
    MRSL_JOB_VGRID,
    MRSL_JOB_EXECUTE,
    MRSL_JOB_EXECUTABLES,
    MRSL_JOB_INPUTFILES,
    MRSL_JOB_OUTPUTFILES,
    MRSL_JOB_NAME,
    MRSL_JOB_RETRIES,
    MRSL_JOB_RUNTIMEENVIROMENT,
    MRSL_JOB_ENVIROMENT,
    MRSL_JOB_VERIFYFILES,
    MRSL_JOB_DISK,
    MRSL_JOB_JOBTYPE,
    MRSL_JOB_RESOURCE,
    MRSL_JOB_CPUTIME,
    MRSL_JOB_PLATFORM,
    MRSL_JOB_USERCERT,
    MRSL_JOB_MAXFILL,
    MRSL_JOB_MOUNT,
    MRSL_JOB_MAXPRICE,
    MRSL_JOB_PROJECT,
    MRSL_JOB_SANDBOX,
    MRSL_JOB_CPUCOUNT,
    MRSL_JOB_NOTIFY,
    MRSL_JOB_NODECOUNT,
    MRSL_JOB_ARCHITECTURE,
    MRSL_JOB_MEMORY
]

SELECTION_START = 'START'
SELECTION_END = 'END'
SELECTION_MAX = 'MAX'
LOWER = 'lower'
UPPER = 'upper'
TOP_BAR = 'top_bar'


# class PollingThread(threading.Thread):
#     """
#     Separate thread used to update the monitor widget by polling data from
#     MiG job queue.
#     """
#     # TODO replace this methodology.
#     def __init__(self, monitor_widget, stop_flag, timer):
#         """
#         Constructor for PollingThread.
#
#         :param monitor_widget: (MonitorWidget) The widget to poll for.
#
#         :param stop_flag: (Event) A flag denoting if the polling should
#         continue or not.
#
#         :param timer: (int) How often to poll, in seconds.
#         """
#         threading.Thread.__init__(self)
#         self.monitor_widget = monitor_widget
#         self.stop_flag = stop_flag
#         self.timer = timer
#
#     def run(self):
#         """
#         Run function for PollingThread, which will start the thread. Until it
#         is flagged to stop this will continue forever, and ask the monitor
#         widget to update periodically, according to the timer given.
#
#         :return: No return.
#         """
#         while not self.stop_flag.wait(3):
#             self.monitor_widget.update_queue_display()


class MonitorWidget:
    """
    Jupyter widget for displaying a MiG job queue. Some basic interactions
    such as cancelling or resubmitting can also be performed. Will update
    periodically.
    """
    def __init__(self, vgrid, timer=60, displayed_jobs=30, debug=True):
        """
        Constructor for MonitorWidget.

        :param vgrid: (str) The VGrid to connect to.

        :param timer: (int)[optional] How often the widget should update. If
        lower than 60, will be changed to 60. Default is 60.

        :param displayed_jobs: (int)[optional] How many jobs to display per
        page. Default is 30.

        :param debug: (bool)[optional] Flag for if the widget is running in
        debug mode. Default value is True.
        """

        self.logfile = create_monitor_logfile(debug)

        check_input(vgrid, str, 'vgrid')
        self.vgrid = vgrid
        check_input(timer, int, 'timer')
        if timer < 60:
            timer = 60
        self.timer = timer

        self.monitor_display_area = widgets.Output()
        self.current_queue_selection = {}
        self.displayed_jobs = displayed_jobs
        self.job_count = 0
        self.jobs = {}
        self.widgets = {}

        # self.__stop_polling = threading.Event()
        self.__start_queue_display(None)

    def __start_queue_display(self, button):
        """
        Starts displaying queue information. Creates and starts a
        PollingThread to keep the display updating automatically.

        :return: No return.
        """
        self.update_queue_display()
        # polling_thread = PollingThread(self, self.__stop_polling, self.timer)
        # polling_thread.daemon = True
        # polling_thread.start()

    def __stop_queue_display(self):
        """
        Sets a flag to stop the PollingThread and prevent an automatic
        update.

        :return: No return.
        """
        # self.__stop_polling.set()

    def update_queue_display(self):
        """
        Sends a request for an up to date job_queue from the vgrid. If a valid
        response is returned then the job display is updated.

        :return: No return.
        """
        #  TODO accommodate this call not working
        valid, result = self.get_vgrid_queue()

        if valid:
            self.jobs = result
            self.__display_job_queue()

    def get_vgrid_queue(self):
        """
        Retrieves a dictionary of jobs from the MiG Vgrid via JSON request.

        :return: (Tuple (bool, str or dict)) If a invalid JSON response is
        received then an tuple of first value False, with a explanatory error
        message as second value is returned. Otherwise, a tuple is returned
        with a first value of True, and with the dictionary of jobs as the
        second value.
        """
        attributes = {}
        _, response, _ = vgrid_job_json_call(
            self.vgrid,
            VGRID_READ,
            VGRID_QUEUE_OBJECT_TYPE,
            attributes,
            logfile=self.logfile
        )

        if response['object_type'] == 'job_dict':
            jobs = response['jobs']
            write_to_log(
                self.logfile,
                'get_vgrid_queue',
                'Got jobs: %s' % jobs.keys())
            return True, jobs
        else:
            msg = 'something went wrong with retrieving the queue. '
            write_to_log(
                self.logfile,
                'get_vgrid_queue',
                '%s: %s' % (msg, response)
            )
            if self.logfile:
                msg += \
                    "You can check the logfile '%s' for details" % self.logfile
            return False, msg

    def __display_job_queue(self, *args):
        """
        Creates a job queue display within the widget. This will display all
        jobs in the queue in the defined range. A top bar of buttons is also
        created.

        :return: No return.
        """
        # first time run through
        if not self.current_queue_selection:
            self.current_queue_selection[SELECTION_MAX] = len(self.jobs)
            self.current_queue_selection[SELECTION_START] = 1

            self.current_queue_selection[SELECTION_END] = self.displayed_jobs
            if len(self.jobs) < self.displayed_jobs:
                self.current_queue_selection[SELECTION_END] = len(self.jobs)
        # got new job selection
        elif self.job_count != len(self.jobs):
            self.current_queue_selection[SELECTION_MAX] = len(self.jobs)
            self.current_queue_selection[SELECTION_START] = \
                self.widgets[LOWER].value
            self.current_queue_selection[SELECTION_END] = \
                self.widgets[UPPER].value
            if len(self.jobs) < self.widgets[UPPER].value:
                self.current_queue_selection[SELECTION_END] = len(self.jobs)
        # any other time
        else:
            self.current_queue_selection[SELECTION_MAX] = len(self.jobs)
            self.current_queue_selection[SELECTION_START] = \
                self.widgets[LOWER].value
            self.current_queue_selection[SELECTION_END] = \
                self.widgets[UPPER].value

        if TOP_BAR not in self.widgets or self.job_count != len(self.jobs):
            lower = widgets.BoundedIntText(
                value=self.current_queue_selection[SELECTION_START],
                min=1,
                max=self.current_queue_selection[SELECTION_END],
                step=1,
                disabled=False,
                layout=widgets.Layout(width='60px')
            )
            lower.observe(self.__display_job_queue, names='value')
            upper = widgets.BoundedIntText(
                value=self.current_queue_selection[SELECTION_END],
                min=self.current_queue_selection[SELECTION_START],
                max=self.current_queue_selection[SELECTION_MAX],
                step=1,
                disabled=False,
                layout=widgets.Layout(width='60px')
            )
            upper.observe(self.__display_job_queue, names='value')

            self.widgets[UPPER] = upper
            self.widgets[LOWER] = lower
            widgets.link((lower, 'value'), (upper, 'min'))
            widgets.link((upper, 'value'), (lower, 'max'))

            top_bar_items = [
                widgets.Label('Displaying '),
                lower,
                widgets.Label(' to '),
                upper,
                widgets.Label(' of %s total jobs for VGrid %s'
                              % (len(self.jobs), self.vgrid))
            ]
            top_bar = widgets.HBox(
                top_bar_items
            )
            self.widgets[TOP_BAR] = top_bar

        grid_items = []
        for _, v in JOB_QUEUE_KEYS.items():
            grid_items.append(
                widgets.Label(
                    value=v,
                    layout=widgets.Layout(width='100%')
                )
            )
        grid_items.append(widgets.Label(''))

        sorted_jobs = sorted(
            self.jobs.items(),
            key=lambda k_v: k_v[1][MRSL_JOB_RECEIVED_TIME],
            reverse=True
        )

        lower_limit = int(self.widgets[LOWER].value) - 1
        if lower_limit < 0:
            lower_limit = 0
        cut_list = sorted_jobs[lower_limit:self.widgets[UPPER].value]
        for job in cut_list:
            grid_items += self.__get_job_display_row(job[1])

        grid = widgets.GridBox(
            grid_items,
            layout=widgets.Layout(
                grid_template_columns="repeat(%s, 25%%)"
                                      % str(len(JOB_QUEUE_KEYS) + 1)
            )
        )
        queue_display = widgets.VBox(
            [
                self.widgets[TOP_BAR],
                grid
            ]
        )

        self.monitor_display_area.clear_output(wait=True)
        with self.monitor_display_area:
            display(queue_display)
        self.job_count = len(self.jobs)

    def __get_job_display_row(self, job):
        """
        Creates a table row displaying a job and its most important details,
        along with buttons to interact with the job directly.

        :param job: (str) The job id for the row.

        :return: (list) A list of all widget elements created for the row.
        """
        row_items = []
        job_id = job[MRSL_JOB_ID]
        for key in JOB_QUEUE_KEYS.keys():
            row_items.append(
                widgets.Label(
                    value=job[key],
                    layout=widgets.Layout(width='100%')
                )
            )
        button_items = []
        for button_dict in self.__get_job_interaction_buttons(job):
            button_args = button_dict['args']
            button_func = button_dict['func']
            button = widgets.Button(
                value=button_args['value'],
                description=button_args['description'],
                button_style=button_args['button_style'],
                tooltip=button_args['tooltip'],
                icon=button_args['icon'],
                disabled=button_args['disabled']
            )

            button.on_click(lambda b, f=button_func: f(b, job_id))
            button_items.append(button)

        buttons = widgets.HBox(button_items)
        row_items.append(buttons)
        return row_items

    def __get_job_interaction_buttons(self, job):
        """
        Creates buttons for interacting with individual jobs. A details,
        resubmit and cancel button are provided. If the job is already
        complete the cancel button is disabled.

        :param job: (str) The job id for these buttons.

        :return: (list) A list of dictionaries defining the three buttons.
        """

        disable_cancel = False
        cancel_tooltip = 'Cancel job'
        if job[MRSL_JOB_STATUS] in ['CANCELED', 'FINISHED', 'FAILED']:
            disable_cancel = True
            cancel_tooltip = 'Cannot cancel job with status  %s' % \
                             job[MRSL_JOB_STATUS].lower()

        return [
            {
                'func': self.__display_func,
                'args': {
                    'value': False,
                    'description': '',
                    'disabled': False,
                    'button_style': '',
                    'tooltip': 'Display job details',
                    'icon': 'plus'
                }
            },
            {
                'func': self.__resubmit_func,
                'args': {
                    'value': False,
                    'description': '',
                    'disabled': False,
                    'button_style': '',
                    'tooltip': 'Resubmit job',
                    'icon': 'refresh'
                }
            },
            {
                'func': self.__cancel_func,
                'args': {
                    'value': False,
                    'description': '',
                    'disabled': disable_cancel,
                    'button_style': '',
                    'tooltip': cancel_tooltip,
                    'icon': 'times'
                }
            },
        ]

    def __display_func(self, button, job_id):
        """
        'Display Job' button clicked event handler. Displays the given jobs
        details in the widget.

        :param button: (widgets.Button) The button object.

        :param job_id: (str) Id of the job to display.

        :return: No return.
        """
        self.__stop_queue_display()

        job = self.jobs[job_id]

        detail_items = [self.__back_to_queue_button()]

        for key, value in JOB_CORE_DISPLAY_KEYS.items():
            detail_items.append(
                widgets.Label("-- %s --" % value)
            )
            msg = ''
            if key in job:
                if isinstance(job[key], list):
                    for elem in job[key]:
                        msg += str(elem)
                else:
                    msg = str(job[key])
            detail_items.append(
                widgets.Label(msg)
            )
        detail_items.append(self.__back_to_queue_button())

        job_details = widgets.VBox(detail_items)

        self.monitor_display_area.clear_output(wait=True)
        with self.monitor_display_area:
            display(job_details)
        self.job_count = len(self.jobs)

    def __back_to_queue_button(self):
        """
        Creates a button to revert view to job queue.

        :return: (widgets.Button) The cancel button.
        """
        button = widgets.Button(
            value=False,
            description='',
            button_style='',
            tooltip='Go back to job queue',
            icon='backward'
        )
        # button.on_click(self.__start_queue_display)
        button.on_click(self.__display_job_queue)
        return button

    def __resubmit_func(self, button, job_id):
        """
        'Resubmit Job' button clicked event handler. Sends a JSON request to
        the MiG VGrid for the job to be ressubmitted.

        :param button: (widgets.Button) The button object.

        :param job_id: (str) The resubmitted job id.

        :return: No return.
        """

        job = self.jobs[job_id]

        attributes = {}
        for key, value in job.items():
            if key in SUBMISSION_JOB_KEYS and value:
                updated_values = []
                if isinstance(value, list):
                    for element in value:
                        if isinstance(element, list):
                            updated_values.append("%s=%s"
                                                  % (element[0], element[1]))
                if updated_values:
                    attributes[key] = updated_values
                else:
                    attributes[key] = value

        _, response, _ = vgrid_job_json_call(
            self.vgrid,
            VGRID_CREATE,
            VGRID_JOB_OBJECT_TYPE,
            attributes
        )

    def __cancel_func(self, button, job_id):
        """
        'Cancel Job' button clicked event handler. Sends a JSON request to
        the MiG VGrid for the job to be cancelled.

        :param button: (widgets.Button) The button object.

        :param job_id: (str) The cancelled job id.

        :return: No return.
        """
        attributes = {
            MRSL_JOB_ID: job_id,
            MRSL_JOB_STATUS: 'CANCELED'
        }
        _, response, _ = vgrid_job_json_call(
            self.vgrid,
            VGRID_UPDATE,
            VGRID_JOB_OBJECT_TYPE,
            attributes
        )

        if OBJECT_TYPE in response and response[OBJECT_TYPE] == 'text':
            self.jobs[job_id][MRSL_JOB_STATUS] = 'CANCELED'

    def display_widget(self):
        """
        Returns the widget in a display ready state.

        :return: (widgets.Output) The output are to be displayed ina  notebook.
        """
        return self.monitor_display_area


def update_monitor(monitor):
    while(True):
        monitor.update_queue_display()
        time.sleep(15)




import os
import json
import shutil
import yaml

from .constants import NAME, PERSISTENCE_ID, INPUT_FILE, TRIGGER_PATHS, \
    RECIPES, OUTPUT, VARIABLES, OBJECT_TYPE, VGRID, TASK_FILE, \
    TRIGGER_RECIPES, SWEEP, DEFAULT_MEOW_IMPORT_EXPORT_DIR, PATTERNS, \
    RECIPE_NAME, PATTERN_NAME
from .meow import Pattern, check_patterns_dict, check_recipes_dict, \
    is_valid_pattern_object
from .validation import valid_pattern_name, dir_exists, valid_dir_path, \
    is_valid_recipe_dict, valid_recipe_name


def write_notebook(source, filename):
    """
    Writes the given notebook source code to a given filename.

    :param source: (dict) The notebook source dictionary.

    :param filename: (str) The filename to write to.

    :return: No return
    """
    with open(filename, 'w') as job_file:
        json.dump(source, job_file)


def write_yaml(source, filename):
    """
    Writes a given objcet to a yaml file.

    :param source: (any) A python object to be written.

    :param filename: (str) The filename to be written to.

    :return: No return
    """
    with open(filename, 'w') as param_file:
        yaml.dump(source, param_file, default_flow_style=False)


def read_yaml(filepath):
    """
    Reads a file path as a yaml object.

    :param filepath: (str) The file to read.

    :return: (object) An object read from the file.
    """
    with open(filepath, 'r') as yaml_file:
        return yaml.full_load(yaml_file)


def make_dir(path, can_exist=True, ensure_clean=False):
    """
    Creates a new directory at the given path.

    :param path: (str) The directory path.

    :param can_exist: (boolean) [optional] A toggle for if a previously
    existing directory at the path will throw an error or not. Default is
    true (e.g. no error is thrown if the path already exists)

    :param ensure_clean: (boolean) [optional] A toggle for if a previously
    existing directory at the path will be replaced with a new emtpy directory.
    Default is False.

    :return: No return
    """
    if not os.path.exists(path):
        os.mkdir(path)
    elif os.path.isfile(path):
        raise ValueError('Cannot make directory in %s as it already '
                         'exists and is a file' % path)
    else:
        if not can_exist:
            if ensure_clean:
                shutil.rmtree(path)
                os.mkdir(path)
            else:
                raise ValueError("Directory %s already exists. " % path)


def patten_to_yaml_dict(pattern):
    """
    Creates a dictionary from a Pattern object to be exported using YAML. This
    hides internal variables that should not be user accessible

    :param pattern: (Pattern) The Pattern object to be exported

    :return: (dict) A dict, expressing the given Pattern
    """
    # Don't export the name, as that will be taken from the file name.
    # Don't export the persistence_id as that is an internal system variable
    # and should not be user editable.
    pattern_yaml = {
        INPUT_FILE: pattern.trigger_file,
        TRIGGER_PATHS: pattern.trigger_paths,
        RECIPES: pattern.recipes,
        OUTPUT: pattern.outputs,
        VARIABLES: pattern.variables,
        SWEEP: pattern.sweep
    }
    return pattern_yaml


def pattern_from_yaml_dict(yaml_dict, name):
    """
    Creates a Pattern object from an imported YAML dict.

    :param yaml_dict: (dict) The imported YAML dict.

    :param name: (str) The name of the imported Pattern.

    :return: (dict) A dict, expressing the given Pattern
    """
    yaml_dict[NAME] = name
    if RECIPES in yaml_dict and TRIGGER_RECIPES not in yaml_dict:
        trigger_id = 'placeholder_id'
        recipe_dict = {
            trigger_id: {}
        }
        for recipe in yaml_dict[RECIPES]:
            recipe_dict[trigger_id][recipe] = {}
        yaml_dict[TRIGGER_RECIPES] = recipe_dict
        yaml_dict.pop(RECIPES)
    pattern = Pattern(yaml_dict)
    return pattern


def recipe_to_yaml_dict(recipe):
    """
    Creates a dictionary from a Recipe dict to be exported using YAML. This
    hides internal variables that should not be user accessible

    :param recipe: (dict) The Recipe dict to be exported

    :return: (dict) A dict, expressing the given Recipe
    """
    recipe_yaml = {}
    for k, v in recipe.items():
        # Don't export the name, as that will be taken from the file name.
        # Don't export the persistence_id, object_type, vgrid or task_file
        # as these are internal system variables and should not be user
        # editable.
        if k not in [NAME, PERSISTENCE_ID, OBJECT_TYPE, VGRID, TASK_FILE]:
            recipe_yaml[k] = v
    return recipe_yaml


def recipe_from_yaml_dict(yaml_dict, name):
    """
    Creates a Recipe dict from an imported YAML dict.

    :param yaml_dict: (dict) The imported YAML dict.

    :param name: (str) The name of the imported Recipe.

    :return: (dict) A dict, expressing the given Recipe
    """
    recipe_dict = {}
    recipe_dict.update(yaml_dict)
    recipe_dict[NAME] = name
    return recipe_dict


def read_dir(directory=DEFAULT_MEOW_IMPORT_EXPORT_DIR):
    """
    Reads in MEOW Patterns and Recipes from yaml files, contained in a local
    directory. This expects there to be two directories within the given
    directory, one containing the Patterns and another containing the Recipes.

    :param directory: (str) The directory to read from. Default is
    'meow_directory'.

    :return: (dict) A dict of Patterns and Recipe objects.
    """
    valid_dir_path(directory, 'directory')
    dir_exists(directory)

    pattern_dir = os.path.join(directory, PATTERNS)
    recipe_dir = os.path.join(directory, RECIPES)

    result = {
        PATTERNS: {},
        RECIPES: {}
    }

    if os.path.exists(pattern_dir):
        pattern_files = [
            f for f in os.listdir(pattern_dir)
            if os.path.isfile(os.path.join(pattern_dir, f))
        ]
        for file_name in pattern_files:
            pattern = read_dir_pattern(
                file_name,
                directory=directory,
                print_errors=True
            )
            result[PATTERNS][file_name] = pattern

    if os.path.exists(recipe_dir):
        recipe_files = [
            f for f in os.listdir(recipe_dir)
            if os.path.isfile(os.path.join(recipe_dir, f))
        ]
        for file_name in recipe_files:
            recipe = read_dir_recipe(
                file_name,
                directory=directory,
                print_errors=True
            )
            result[RECIPES][file_name] = recipe

    return result


def write_dir(patterns, recipes, directory=DEFAULT_MEOW_IMPORT_EXPORT_DIR):
    """
    Saves the given patterns and recipes in the given directory.

    :param patterns: (dict) A dict of Pattern objects

    :param recipes: (dict) A dict of Recipe dictionaries

    :param directory: (str) The directory to save the object in. Default is
    'meow_directory'

    :return: (No return)
    """
    valid, feedback = check_patterns_dict(patterns, integrity=True)
    if not valid:
        raise ValueError(feedback)

    valid, feedback = check_recipes_dict(recipes)
    if not valid:
        raise ValueError(feedback)

    dir_exists(directory, create=True)

    for pattern in patterns.values():
        write_dir_pattern(pattern, directory=directory)

    for recipe in recipes.values():
        write_dir_recipe(recipe, directory=directory)


def read_dir_pattern(pattern_name, directory=DEFAULT_MEOW_IMPORT_EXPORT_DIR,
                     print_errors=False):
    """
    Read a specific Pattern within the given local directory. There should be
    an intermediate directory, 'Patterns' between the two.

    :param pattern_name: (str) the pattern file to read.

    :param directory: (str) a local directory to read from. Default is
    'meow_directory'.

    :param print_errors: (bool) [Optional] Toggle for if encountered errors
    result in a print statement or throwing an exception. Default is to throw
    an exception.

    :return: (Pattern) The read in pattern object.
    """
    valid_pattern_name(pattern_name)
    valid_dir_path(directory, 'directory')
    dir_exists(directory)

    pattern_dir = os.path.join(directory, PATTERNS)
    dir_exists(pattern_dir)

    try:
        pattern_yaml_dict = read_yaml(os.path.join(pattern_dir, pattern_name))
        if '.' in pattern_name:
            pattern_name = pattern_name[:pattern_name.index('.')]

        pattern = \
            pattern_from_yaml_dict(pattern_yaml_dict, pattern_name)

        return pattern
    except Exception as ex:
        msg = "Tried to import %s '%s', but could not. %s" \
              % (PATTERN_NAME, pattern_name, ex)
        if print_errors:
            print(msg)
        else:
            raise Exception(msg)


def read_dir_recipe(recipe_name, directory=DEFAULT_MEOW_IMPORT_EXPORT_DIR,
                    print_errors=False):
    """
    Read a specific recipe within the given local directory. There should be
    an intermediate directory, 'Recipes' between the two.

    :param recipe_name: (str) the recipe file to read.

    :param directory: (str) a local directory to read from. Default is
    'meow_directory'.

    :param print_errors: (bool) Toggle for if encountered errors result in a
    print statement or throwing an exception. Default is to throw an exception.

    :return: (dict) The read in recipe dict.
    """
    valid_recipe_name(recipe_name)
    valid_dir_path(directory, 'directory')
    dir_exists(directory)

    recipe_dir = os.path.join(directory, RECIPES)
    dir_exists(recipe_dir)

    try:

        recipe_yaml_dict = read_yaml(os.path.join(recipe_dir, recipe_name))
        if '.' in recipe_name:
            recipe_name = recipe_name[:recipe_name.index('.')]

        recipe = recipe_from_yaml_dict(recipe_yaml_dict, recipe_name)

        return recipe
    except Exception as ex:
        msg = "Tried to import %s '%s', but could not. %s" \
              % (RECIPE_NAME, recipe_name, ex)
        if print_errors:
            print(msg)
        else:
            raise Exception(msg)


def write_dir_pattern(pattern, directory=DEFAULT_MEOW_IMPORT_EXPORT_DIR):
    """
    Saves a given pattern locally.

    :param pattern: (Pattern) the pattern to save.

    :param directory: (str) The directory to write the Pattern to.

    :return: (No return)
    """
    valid, feedback = is_valid_pattern_object(pattern, integrity=True)

    if not valid:
        msg = "Could not export %s %s. %s" \
              % (PATTERN_NAME, pattern.name, feedback)
        raise ValueError(msg)

    dir_exists(directory, create=True)
    pattern_dir = os.path.join(directory, PATTERNS)
    dir_exists(pattern_dir, create=True)

    pattern_file_path = os.path.join(pattern_dir, pattern.name)
    pattern_yaml = patten_to_yaml_dict(pattern)

    write_yaml(pattern_yaml, pattern_file_path)


def write_dir_recipe(recipe, directory=DEFAULT_MEOW_IMPORT_EXPORT_DIR):
    """
    Saves a given recipe locally.

    :param recipe: (dict) the recipe dict to save.

    :param directory: (str) The directory to write the Recipe to.

    :return: (No return)
    """
    valid, feedback = is_valid_recipe_dict(recipe)
    dir_exists(directory, create=True)

    if not valid:
        msg = "Could not export %s %s. %s" \
              % (RECIPE_NAME, recipe['NAME'], feedback)
        raise ValueError(msg)

    recipe_dir = os.path.join(directory, RECIPES)
    dir_exists(recipe_dir, create=True)

    recipe_file_path = os.path.join(recipe_dir, recipe[NAME])
    recipe_yaml = recipe_to_yaml_dict(recipe)

    write_yaml(recipe_yaml, recipe_file_path)


def delete_dir_pattern(pattern_name, directory=DEFAULT_MEOW_IMPORT_EXPORT_DIR):
    """
    Removes a a saved pattern by the given name.

    :param pattern_name: (str or Pattern) Name of pattern to delete, or
    complete Pattern object to delete.

    :param directory: (str) Directory containing pattern saves. Default is
    'meow_directory'

    :return: (No return)
    """
    if isinstance(pattern_name, Pattern):
        pattern_name = Pattern.name
    if not isinstance(pattern_name, str):
        raise ValueError("'pattern_name' must be either a string or a Pattern")
    valid_pattern_name(pattern_name)
    dir_exists(directory)

    pattern_dir = os.path.join(directory, PATTERNS)
    dir_exists(pattern_dir)

    file_path = os.path.join(pattern_dir, pattern_name)
    if os.path.exists(file_path):
        os.remove(file_path)


def delete_dir_recipe(recipe_name, directory=DEFAULT_MEOW_IMPORT_EXPORT_DIR):
    """
    Removes a a saved recipe by the given name.

    :param recipe_name: (str or dict) Name of recipe to delete, or a recipe
    dict to delete.

    :param directory: (str) Directory containing recipe saves. Default is
    'meow_directory'

    :return: (No return)
    """
    if isinstance(recipe_name, dict):
        recipe_name = recipe_name[NAME]
    if not isinstance(recipe_name, str):
        raise ValueError("'recipe_name' must be either a string or a dict")
    valid_recipe_name(recipe_name)
    dir_exists(directory)

    recipe_dir = os.path.join(directory, RECIPES)
    dir_exists(recipe_dir)

    file_path = os.path.join(recipe_dir, recipe_name)
    if os.path.exists(file_path):
        os.remove(file_path)



import requests
import json
import traceback
import os

from .constants import VGRID_PATTERN_OBJECT_TYPE, VGRID_RECIPE_OBJECT_TYPE, \
    NAME, INPUT_FILE, TRIGGER_PATHS, OUTPUT, RECIPES, VARIABLES, \
    VGRID_CREATE, VGRID, VGRID_READ, VGRID_DELETE, VGRID_ANY_OBJECT_TYPE, \
    VALID_OPERATIONS, VALID_WORKFLOW_TYPES, VALID_JOB_TYPES, \
    DEFAULT_JSON_TIMEOUT, PATTERNS, VGRID_WORKFLOWS_OBJECT, VGRID_TEXT_TYPE, \
    OBJECT_TYPE, VGRID_ERROR_TYPE, RECIPE_NAME, RECIPE, SOURCE, VGRID_UPDATE, \
    PERSISTENCE_ID, PATTERN_NAME, SWEEP
from .validation import check_input, valid_recipe_name, is_valid_recipe_dict, \
    valid_pattern_name
from .logging import write_to_log
from .meow import Pattern, is_valid_pattern_object


MRSL_VGRID = 'VGRID'


def export_pattern_to_vgrid(vgrid, pattern, timeout=DEFAULT_JSON_TIMEOUT):
    """
    Exports a given pattern to a MiG based Vgrid. Raises a TypeError or
    ValueError if the pattern is not valid. Note this function is not used
    within mig_meow and is intended for users who want to programmatically
    alter vgrid workflows.

    :param vgrid: (str) Vgrid to which pattern will be exported.

    :param pattern: (Pattern) Pattern object to export.

    :param timeout: (int) [optional] Timeout duration in seconds for Vgrid
    call. Default is 60

    :return: (function call to vgrid_workflow_json_call) if pattern is valid,
    will call function 'vgrid_workflow_json_call'.
    """
    check_input(vgrid, str, 'vgrid')

    if not isinstance(pattern, Pattern):
        raise TypeError(
            "The provided object '%s' is a %s, not a Pattern as expected"
            % (pattern, type(pattern))
        )
    status, msg = pattern.integrity_check()
    if not status:
        raise ValueError(
            'The provided pattern is not a valid Pattern. %s' % msg
        )

    attributes = {
        NAME: pattern.name,
        INPUT_FILE: pattern.trigger_file,
        TRIGGER_PATHS: pattern.trigger_paths,
        OUTPUT: pattern.outputs,
        RECIPES: pattern.recipes,
        VARIABLES: pattern.variables
    }
    return vgrid_workflow_json_call(
        vgrid,
        VGRID_CREATE,
        VGRID_PATTERN_OBJECT_TYPE,
        attributes,
        timeout=timeout
    )


def export_recipe_to_vgrid(vgrid, recipe, timeout=DEFAULT_JSON_TIMEOUT):
    """
    Exports a given recipe to a MiG based Vgrid. Raises a TypeError or
    ValueError if the recipe is not valid. Note this function is not used
    within mig_meow and is intended for users who want to programmatically
    alter vgrid workflows.

    :param vgrid: (str) Vgrid to which recipe will be exported.

    :param recipe: (dict) Recipe object to export.

    :param timeout: (int) [optional] Timeout duration in seconds for Vgrid
    call. Default is 60

    :return: (function call to vgrid_workflow_json_call) if recipe is valid,
    will call function 'vgrid_workflow_json_call'.
    """
    check_input(vgrid, str, 'vgrid')

    if not isinstance(recipe, dict):
        raise TypeError("The provided object '%s' is a %s, not a dict "
                        "as expected" % (recipe, type(recipe)))
    status, msg = is_valid_recipe_dict(recipe)
    if not status:
        raise ValueError('The provided recipe is not valid. %s' % msg)

    return vgrid_workflow_json_call(
        vgrid,
        VGRID_CREATE,
        VGRID_RECIPE_OBJECT_TYPE,
        recipe,
        timeout=timeout
    )


def vgrid_workflow_json_call(
        vgrid, operation, workflow_type, attributes, logfile=None,
        ssl=True, timeout=DEFAULT_JSON_TIMEOUT):
    """
    Validates input for a JSON workflow call to VGRID. Raises a TypeError or
    ValueError if an invalid value is found. If no problems are found then a
    JSON message is setup.

    :param vgrid: (str) Vgrid to which workflow will be exported.

    :param operation: (str) The operation type to be performed by the MiG based
    JSON API. Valid operations are 'create', 'read', 'update' and 'delete'.

    :param workflow_type: (str) MiG workflow object type. Valid are
    'workflows', 'workflowpattern', 'workflowrecipe', and 'any',

    :param attributes: (dict) A dictionary of arguments defining the specifics
    of the requested operation.

    :param timeout: (int) [optional] Timeout duration in seconds for Vgrid
    call. Default is 60

    :param logfile: (str)[optional] Path to a logfile. If provided logs are
    recorded in this file. Default is None.

    :param ssl: (boolean)[optional] Toggle to use ssl checks. Default True

    :return: (function call to __vgrid_json_call) If all inputs are valid,
    will call function '__vgrid_json_call'.
    """
    check_input(vgrid, str, 'vgrid')
    check_input(operation, str, 'operation')
    check_input(workflow_type, str, 'workflow_type')
    check_input(attributes, dict, 'attributes', or_none=True)
    check_input(ssl, bool, 'ssl')

    try:
        url = os.environ['WORKFLOWS_URL']
    except KeyError:
        msg = \
            'MiGrid WORKFLOWS_URL was not specified in the local ' \
            'environment. This should be created automatically as part of ' \
            'the Notebook creation if the Notebook was created on IDMC. ' \
            'Currently this is the only supported way to interact with a ' \
            'VGrid. '
        write_to_log(logfile, '__vgrid_json_call', msg)
        raise EnvironmentError(msg)

    write_to_log(
        logfile,
        'vgrid_workflow_json_call',
        'A vgrid call has been requested. vgrid=%s, workflow_type=%s, '
        'attributes=%s' % (vgrid, workflow_type, attributes)
    )

    if operation not in VALID_OPERATIONS:
        msg = \
            'Requested operation %s is not a valid operation. Valid ' \
            'operations are: %s' % (operation, VALID_OPERATIONS)
        write_to_log(logfile, 'vgrid_workflow_json_call', msg)
        raise ValueError(msg)

    if workflow_type not in VALID_WORKFLOW_TYPES:
        msg = \
            'Requested workflow type %s is not a valid workflow type. Valid ' \
            'workflow types are: %s' % (workflow_type, VALID_WORKFLOW_TYPES)
        write_to_log(logfile, 'vgrid_workflow_json_call', msg)
        raise ValueError(msg)

    attributes[VGRID] = vgrid

    return __vgrid_json_call(
        operation,
        workflow_type,
        attributes,
        url,
        logfile=logfile,
        ssl=ssl,
        timeout=timeout
    )


def vgrid_job_json_call(
        vgrid, operation, workflow_type, attributes, logfile=None,
        ssl=True, timeout=DEFAULT_JSON_TIMEOUT):
    """
    Validates input for a JSON job call to VGRID. Raises a TypeError or
    ValueError if an invalid value is found. If no problems are found then a
    JSON message is setup.

    :param vgrid: (str) Vgrid to which recipe will be exported.

    :param operation: (str) The operation type to be performed by the MiG based
    JSON API. Valid operations are 'create', 'read', 'update' and 'delete'.

    :param workflow_type: (str) MiG workflow action type. Valid are
    'queue', 'job', 'cancel_job', and 'resubmit_job',

    :param attributes: (dict) A dictionary of arguments defining the specifics
    of the requested operation.

    :param logfile: (str)[optional] Path to a logfile. If provided logs are
    recorded in this file. Default is None.

    :param ssl: (boolean)[optional] Toggle to skip ssl checks. Default
    False

    :param timeout: (int) [optional] Timeout duration in seconds for Vgrid
    call. Default is 60

    :return: (function call to __vgrid_json_call) If all inputs are valid,
    will call function '__vgrid_json_call'.
    """
    check_input(vgrid, str, 'vgrid')
    check_input(operation, str, 'operation')
    check_input(workflow_type, str, 'workflow_type')
    check_input(attributes, dict, 'attributes', or_none=True)
    check_input(ssl, bool, 'ssl')

    try:
        url = os.environ['JOBS_URL']
    except KeyError:
        msg = \
            'MiGrid JOBS_URL was not specified in the local ' \
            'environment. This should be created automatically as part of ' \
            'the Notebook creation if the Notebook was created on IDMC. ' \
            'Currently this is the only supported way to interact with a ' \
            'VGrid. '
        write_to_log(logfile, '__vgrid_json_call', msg)
        raise EnvironmentError(msg)

    write_to_log(
        logfile,
        'vgrid_job_json_call',
        'A vgrid call has been requested. vgrid=%s, workflow_type=%s, '
        'attributes=%s' % (vgrid, workflow_type, attributes)
    )

    if operation not in VALID_OPERATIONS:
        msg = \
            'Requested operation %s is not a valid operation. Valid ' \
            'operations are: %s' % (operation, VALID_OPERATIONS)
        write_to_log(logfile, 'vgrid_job_json_call',  msg)
        raise ValueError(msg)

    if workflow_type not in VALID_JOB_TYPES:
        msg = \
            'Requested workflow type %s is not a valid workflow type. Valid ' \
            'workflow types are: %s' % (workflow_type, VALID_JOB_TYPES)
        write_to_log(logfile, 'vgrid_job_json_call', msg)
        raise ValueError(msg)

    attributes[VGRID] = vgrid

    return __vgrid_json_call(
        operation,
        workflow_type,
        attributes,
        url,
        logfile=logfile,
        ssl=ssl,
        timeout=timeout
    )


def __vgrid_json_call(
        operation, workflow_type, attributes, url, logfile=None, ssl=True,
        timeout=DEFAULT_JSON_TIMEOUT):
    """
    Makes JSON call to MiG. Will pull url and session_id from local
    environment variables, as setup by MiG notebook spawner. Will raise
    EnvironmentError if these are not present.

    :param operation: (str) The operation type to be performed by the MiG based
    JSON API. Valid operations are 'create', 'read', 'update' and 'delete'.

    :param workflow_type: (str) MiG workflow action type. Valid are
    'workflows', 'workflowpattern', 'workflowrecipe', 'any', 'queue', 'job',
    'cancel_job', and 'resubmit_job'

    :param attributes: (dict) A dictionary of arguments defining the specifics
    of the requested operation.

    :param url: (str) The url to send JSON call to.

    :param logfile: (str)[optional] Path to a logfile. If provided logs are
    recorded in this file. Default is None.

    :param ssl: (boolean)[optional] Toggle to use ssl checks. Default True

    :param timeout: (int) [optional] Timeout duration in seconds for Vgrid
    call. Default is 60

    :return: (Tuple (dict, dict, dict) Returns JSON call results as three
    dicts. First is the header, then the body then the footer. Header contains
    keys 'headers' and 'object_type', Body contains 'workflows' and
    'object_type' and footer contains 'text' and 'object_type'.
    """

    try:
        session_id = os.environ['WORKFLOWS_SESSION_ID']
    except KeyError:
        msg = \
            'MiGrid WORKFLOWS_SESSION_ID was not specified in the local ' \
            'environment. This should be created automatically as part of ' \
            'the Notebook creation if the Notebook was created on IDMC. ' \
            'Currently this is the only supported way to interact with a ' \
            'VGrid. '
        write_to_log(logfile, '__vgrid_json_call', msg)
        raise EnvironmentError(msg)

    data = {
        'workflowsessionid': session_id,
        'operation': operation,
        'type': workflow_type,
        'attributes': attributes
    }

    write_to_log(
        logfile,
        '__vgrid_json_call',
        'sending request to  %s with data: %s' % (url, data)
    )

    try:
        response = requests.post(
            url,
            json=data,
            verify=ssl,
            timeout=timeout
        )

    except requests.Timeout:
        msg = 'Connection to MiG has timed out. '
        write_to_log(logfile, '__vgrid_json_call', msg)
        write_to_log(logfile, '__vgrid_json_call', traceback.format_exc())
        msg += 'Please check that the MiG is still online. If the problem ' \
               'persists contact an admin. '
        raise Exception(msg)
    except requests.ConnectionError:
        msg = 'Connection could not be established. '
        write_to_log(logfile, '__vgrid_json_call', msg)
        write_to_log(logfile, '__vgrid_json_call', traceback.format_exc())
        msg += 'Please check that the MiG is still online. If the problem ' \
               'persists contact an admin. '
        raise Exception(msg)

    try:
        json_response = response.json()

    except json.JSONDecodeError as err:
        msg = 'Unexpected feedback from MiG. %s' % err
        write_to_log(logfile, '__vgrid_json_call', msg)
        raise Exception(msg)

    header = json_response[0]
    body = json_response[1]
    footer = json_response[2]

    return header, body, footer


def read_vgrid(vgrid, ssl=True, timeout=DEFAULT_JSON_TIMEOUT):
    """
    Reads a given vgrid. Returns a dict of Patterns and a dict of Recipes .

    :param vgrid: (str) A vgrid to read

    :param ssl: (boolean)[optional] Toggle to use ssl checks. Default True

    :param timeout: (int) [optional] Timeout duration in seconds for Vgrid
    call. Default is 60

    :return: (dict) A dictionary of responses. Contains separate keys for the
    patterns and the recipes.
    """

    check_input(vgrid, str, VGRID)

    _, response, _ = vgrid_workflow_json_call(
        vgrid,
        VGRID_READ,
        VGRID_ANY_OBJECT_TYPE,
        {},
        ssl=ssl,
        timeout=timeout
    )

    output = {
        PATTERNS: {},
        RECIPES: {}
    }
    response_patterns = {}
    response_recipes = {}
    if VGRID_WORKFLOWS_OBJECT in response:
        for response_object in response[VGRID_WORKFLOWS_OBJECT]:
            if response_object[OBJECT_TYPE] == VGRID_PATTERN_OBJECT_TYPE:
                response_patterns[response_object[NAME]] = \
                    Pattern(response_object)
            elif response_object[OBJECT_TYPE] == VGRID_RECIPE_OBJECT_TYPE:
                response_recipes[response_object[NAME]] = response_object

        output[PATTERNS] = response_patterns
        output[RECIPES] = response_recipes
        return output

    elif OBJECT_TYPE in response and response[OBJECT_TYPE] == VGRID_ERROR_TYPE:
        print("Could not retrieve workflow objects. %s"
              % response[VGRID_TEXT_TYPE])
        return output
    else:
        print("Unexpected response: {}".format(response))
        return output


def write_vgrid(
        patterns, recipes, vgrid, ssl=True, timeout=DEFAULT_JSON_TIMEOUT):
    """
    Writes a collection of patterns and recipes to a given vgrid.

    :param patterns: (dict) A dictionary of pattern objects.

    :param recipes: (recipes) A dictionary of recipes.

    :param vgrid: (str) The vgrid to write to.

    :param ssl: (boolean)[optional] Toggle to use ssl checks. Default True

    :param timeout: (int) [optional] Timeout duration in seconds for Vgrid
    call. Default is 60

    :return: (dict) Dicts of updated patterns and recipes.
    """
    check_input(vgrid, str, VGRID)
    check_input(patterns, dict, 'patterns', or_none=True)
    check_input(recipes, dict, 'recipes', or_none=True)

    updated_patterns = {}
    for pattern in patterns.values():
        new_pattern = write_vgrid_pattern(
            pattern,
            vgrid,
            ssl=ssl,
            timeout=timeout
        )
        updated_patterns[pattern.name] = new_pattern

    updated_recipes = {}
    for recipe in recipes.values():
        new_recipe = write_vgrid_recipe(
            recipe,
            vgrid,
            ssl=ssl,
            timeout=timeout
        )
        updated_recipes[recipe[NAME]] = new_recipe

    return {
        PATTERNS: updated_patterns,
        RECIPES: updated_recipes
    }


def read_vgrid_pattern(pattern, vgrid, ssl=True, timeout=DEFAULT_JSON_TIMEOUT):
    """
    Reads a given pattern from a given vgrid.

    :param pattern: (str) The pattern name to read.

    :param vgrid: (str) The Vgrid to read from

    :param ssl: (boolean)[optional] Toggle to use ssl checks. Default True

    :param timeout: (int) [optional] Timeout duration in seconds for Vgrid
    call. Default is 60

    :return: (Pattern) A pattern object, or None if a Pattern could not be
    found
    """
    check_input(vgrid, str, VGRID)
    valid_pattern_name(pattern)

    attributes = {
        NAME: pattern
    }

    _, response, _ = vgrid_workflow_json_call(
        vgrid,
        VGRID_READ,
        VGRID_PATTERN_OBJECT_TYPE,
        attributes,
        ssl=ssl,
        timeout=timeout
    )

    if OBJECT_TYPE in response \
            and response[OBJECT_TYPE] == VGRID_WORKFLOWS_OBJECT \
            and VGRID_WORKFLOWS_OBJECT in response:
        pattern_list = response[VGRID_WORKFLOWS_OBJECT]
        if len(pattern_list) > 1:
            print("Got several matching %ss: %s"
                  % (PATTERN_NAME, [entry[NAME] for entry in pattern_list]))
        return Pattern(pattern_list[0])
    elif OBJECT_TYPE in response and response[OBJECT_TYPE] == VGRID_ERROR_TYPE:
        print("Could not retrieve workflow %s. %s"
              % (PATTERN_NAME, response[VGRID_TEXT_TYPE]))
        return None
    else:
        print("Got unexpected response. %s" % response)
        return None


def read_vgrid_recipe(recipe, vgrid, ssl=True, timeout=DEFAULT_JSON_TIMEOUT):
    """
    Reads a given recipe from a given vgrid.

    :param recipe: (str) The recipe name to read.

    :param vgrid: (str) The Vgrid to read from

    :param ssl: (boolean)[optional] Toggle to use ssl checks. Default True

    :param timeout: (int) [optional] Timeout duration in seconds for Vgrid
    call. Default is 60

    :return: (dict) A recipe dict, or None if a recipe could not be found
    """
    check_input(vgrid, str, VGRID)
    valid_recipe_name(recipe)

    attributes = {
        NAME: recipe
    }

    _, response, _ = vgrid_workflow_json_call(
        vgrid,
        VGRID_READ,
        VGRID_RECIPE_OBJECT_TYPE,
        attributes,
        ssl=ssl,
        timeout=timeout
    )

    if OBJECT_TYPE in response \
            and response[OBJECT_TYPE] == VGRID_WORKFLOWS_OBJECT \
            and VGRID_WORKFLOWS_OBJECT in response:
        recipe_list = response[VGRID_WORKFLOWS_OBJECT]
        if len(recipe_list) > 1:
            print("Got several matching %ss: %s"
                  % (RECIPE_NAME, [entry[NAME] for entry in recipe_list]))
        return recipe_list[0]
    elif OBJECT_TYPE in response and response[OBJECT_TYPE] == VGRID_ERROR_TYPE:
        print("Could not retrieve workflow %s. %s"
              % (RECIPE_NAME, response[VGRID_TEXT_TYPE]))
        return None
    else:
        print("Got unexpected response. %s" % response)
        return None


def write_vgrid_pattern(
        pattern, vgrid, ssl=True, timeout=DEFAULT_JSON_TIMEOUT):
    """
    Creates a new Pattern on a given VGrid, or updates an existing Pattern.

    :param pattern: (Pattern) The pattern object to write to the VGrid.

    :param vgrid: (str) The vgrid to write the pattern to.

    :param ssl: (boolean)[optional] Toggle to use ssl checks. Default True

    :param timeout: (int) [optional] Timeout duration in seconds for Vgrid
    call. Default is 60

    :return: (Pattern) The registered Pattern object.
    """

    check_input(vgrid, str, VGRID)
    is_valid_pattern_object(pattern)

    attributes = {
        NAME: pattern.name,
        INPUT_FILE: pattern.trigger_file,
        TRIGGER_PATHS: pattern.trigger_paths,
        OUTPUT: pattern.outputs,
        RECIPES: pattern.recipes,
        VARIABLES: pattern.variables,
        SWEEP: pattern.sweep
    }

    if hasattr(pattern, 'persistence_id'):
        attributes[PERSISTENCE_ID] = pattern.persistence_id,
        operation = VGRID_UPDATE
    else:
        operation = VGRID_CREATE

    _, response, _ = vgrid_workflow_json_call(
        vgrid,
        operation,
        VGRID_PATTERN_OBJECT_TYPE,
        attributes,
        ssl=ssl,
        timeout=timeout
    )

    if response['object_type'] != 'error_text':
        if operation == VGRID_UPDATE:
            print("%s '%s' updated on VGrid '%s'"
                  % (PATTERN_NAME, pattern.name, vgrid))
        else:
            pattern.persistence_id = response['text']
            print("%s '%s' created on VGrid '%s'"
                  % (PATTERN_NAME, pattern.name, vgrid))
    else:
        if hasattr(pattern, 'persistence_id'):
            delattr(pattern, 'persistence_id')
        print(response['text'])
    return pattern


def write_vgrid_recipe(recipe, vgrid, ssl=True, timeout=DEFAULT_JSON_TIMEOUT):
    """
    Creates a new recipe on a given VGrid, or updates an existing recipe.

    :param recipe: (dict) The recipe to write to the VGrid.

    :param vgrid: (str) The vgrid to write the recipe to.

    :param ssl: (boolean)[optional] Toggle to use ssl checks. Default True

    :param timeout: (int) [optional] Timeout duration in seconds for Vgrid
    call. Default is 60

    :return: (dict) The registered Recipe dict.
    """

    check_input(vgrid, str, VGRID)
    is_valid_recipe_dict(recipe)

    attributes = {
        NAME: recipe[NAME],
        RECIPE: recipe[RECIPE],
        SOURCE: recipe[SOURCE]
    }

    if PERSISTENCE_ID in recipe:
        attributes[PERSISTENCE_ID] = recipe[PERSISTENCE_ID],
        operation = VGRID_UPDATE
    else:
        operation = VGRID_CREATE

    _, response, _ = vgrid_workflow_json_call(
        vgrid,
        operation,
        VGRID_RECIPE_OBJECT_TYPE,
        attributes,
        ssl=ssl,
        timeout=timeout
    )

    if response['object_type'] != 'error_text':
        if operation == VGRID_UPDATE:
            print("%s '%s' updated on VGrid '%s'"
                  % (RECIPE_NAME, recipe[NAME], vgrid))
        else:
            recipe[PERSISTENCE_ID] = response['text']
            print("%s '%s' created on VGrid '%s'"
                  % (RECIPE_NAME, recipe[NAME], vgrid))
    else:
        if PERSISTENCE_ID in recipe:
            recipe.pop(PERSISTENCE_ID)
        print(response['text'])
    return recipe


def delete_vgrid_pattern(
        pattern, vgrid, ssl=True, timeout=DEFAULT_JSON_TIMEOUT):
    """
    Attempts to delete a given pattern from a given VGrid.

    :param pattern: (Pattern) A valid workflow Pattern object

    :param vgrid: (str) A MiG Vgrid to connect to

    :param ssl: (boolean)[optional] Toggle to use ssl checks. Default True

    :param timeout: (int) [optional] Timeout duration in seconds for Vgrid
    call. Default is 60

    :return: (dict) Returns a Pattern object. If the deletion is successful
    the persistence_id attribute is removed
    """

    check_input(vgrid, str, VGRID)
    is_valid_pattern_object(pattern)

    try:
        attributes = {
            PERSISTENCE_ID: pattern.persistence_id,
            NAME: pattern.name
        }
    except AttributeError:
        msg = "Cannot delete a %s that has not been previously registered " \
              "with the VGrid. If you have registered this %s with a Vgrid " \
              "before, then please re-read it again, as necessary data has " \
              "been lost. " % (PATTERN_NAME, PATTERN_NAME)
        print(msg)
        return pattern

    _, response, _ = vgrid_workflow_json_call(
        vgrid,
        VGRID_DELETE,
        VGRID_PATTERN_OBJECT_TYPE,
        attributes,
        ssl=ssl,
        timeout=timeout
    )

    if response['object_type'] != 'error_text':
        delattr(pattern, 'persistence_id')
        print("%s '%s' deleted from VGrid '%s'"
              % (PATTERN_NAME, pattern.name, vgrid))
    else:
        print(response['text'])
    return pattern


def delete_vgrid_recipe(recipe, vgrid, ssl=True, timeout=DEFAULT_JSON_TIMEOUT):
    """
    Attempts to delete a given recipe from a given VGrid.

    :param recipe: (dict) A valid workflow recipe

    :param vgrid: (str) A MiG Vgrid to connect to

    :param ssl: (boolean)[optional] Toggle to use ssl checks. Default True

    :param timeout: (int) [optional] Timeout duration in seconds for Vgrid
    call. Default is 60

    :return: (dict) Returns a recipe dictionary. If the deletion is successful
    the persistence_id attribute is removed
    """

    check_input(vgrid, str, VGRID)
    is_valid_recipe_dict(recipe)

    if PERSISTENCE_ID not in recipe:
        msg = "Cannot delete a %s that has not been previously registered " \
              "with the VGrid. If you have registered this %s with a Vgrid " \
              "before, then please re-read it again, as necessary data has " \
              "been lost. " % (RECIPE_NAME, RECIPE_NAME)
        print(msg)
        return recipe
    attributes = {
        PERSISTENCE_ID: recipe[PERSISTENCE_ID],
        NAME: recipe[NAME]
    }

    _, response, _ = vgrid_workflow_json_call(
        vgrid,
        VGRID_DELETE,
        VGRID_RECIPE_OBJECT_TYPE,
        attributes,
        ssl=ssl,
        timeout=timeout
    )

    if response['object_type'] != 'error_text':
        recipe.pop(PERSISTENCE_ID)
        print("%s '%s' deleted from VGrid '%s'"
              % (RECIPE_NAME, recipe[NAME], vgrid))
    else:
        print(response['text'])
    return recipe




import os
import threading

from datetime import datetime

from .constants import LOGGING_DIR, WORKFLOW_LOGFILE_NAME, \
    MONITOR_LOGFILE_NAME, RUNNER_LOGFILE_NAME

lock = threading.Lock()


def __create_logfile(mode, title):
    """
    Creates a new logfile within the logging directory.

    :param mode: (boolean) Flag for if logfile is required. If false then a
    log is not created and None is returned.

    :param title: (str) The logfile to be created.

    :return: (str or None) If the logfile is created the path of said file is
    returned. If it could not be created then None is returned.
    """
    # TODO improve this

    if not mode:
        return False

    try:
        time = str(datetime.now())
        if not os.path.exists(LOGGING_DIR):
            os.mkdir(LOGGING_DIR)
        log_filename = \
            os.path.join(LOGGING_DIR, "%s_%s.log" % (time, title))
        with open(log_filename, 'w') as logfile:
            logfile.write('Start of %s log at %s\n' % (title, time))
        return log_filename
    except Exception:
        return None


def create_workflow_logfile(debug_mode=None):
    """
    Creates a new logfile for a workflow widget.

    :param debug_mode: (boolean)[optional] flag for widget debug mode. If True
    then a log is created. If false it is not. Default is None.

    :return: (function call to __create_logfile)
    """
    return __create_logfile(debug_mode, WORKFLOW_LOGFILE_NAME)


def create_monitor_logfile(debug_mode=None):
    """
    Creates a new logfile for a monitor widget.

    :param debug_mode: (boolean)[optional] flag for widget debug mode. If True
    then a log is created. If false it is not. Default is None.

    :return: (function call to __create_logfile)
    """
    return __create_logfile(debug_mode, MONITOR_LOGFILE_NAME)


def create_localrunner_logfile(debug_mode=None):
    """
    Creates a new logfile for a local runner.

    :param debug_mode: (boolean)[optional] flag for runner debug mode. If True
    then a log is created. If false it is not. Default is None.

    :return: (function call to __create_logfile)
    """
    return __create_logfile(debug_mode, RUNNER_LOGFILE_NAME)


def write_to_log(log, anchor, entry, to_print=False):
    """
    Writes a new entry to a logfile.

    :param log: (str or None) Path to a logfile to write in. If no log is
    provided because the widget is not in debug mode then the entry is not
    written.

    :param anchor: (str) A string to help locate where this log message
    originated from. Should usually be the name of the function calling this
    function.

    :param entry: (str) Line to write to logfile.

    :param to_print: (boolean) Also print entry. Default is False

    :return: No return.
    """
    if log:
        lock.acquire()
        time = str(datetime.now())
        with open(log, 'a') as logfile:
            logfile.write("%s: %s - %s\n" % (time, anchor, entry))
        lock.release()
    if to_print:
        print("%s - %s" % (anchor, entry))



from .cwl import *
from .info import *
from .meow import *
from .mig import *
from .monitor_widget import *
from .notebook import *
from .workflow_widget import *
from .fileio import *
from.version import __name__, __fullname__, __version__
